import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from pandas.tseries.frequencies import to_offset
from typing import Tuple, Optional, Callable, Any


class TimeSeries:
    """
        A TimeSeries is an immutable object defined by the following three components.

        :param series: The actual time series, as a pandas Series with a proper time index.
        :param confidence_lo: Optionally, a Pandas Series representing lower confidence interval.
        :param confidence_hi: Optionally, a Pandas Series representing upper confidence interval.

        Within this class, TimeSeries type annotations are 'TimeSeries'; see:
        https://stackoverflow.com/questions/15853469/putting-current-class-as-return-type-annotation
    """
    def __init__(self, series: pd.Series, confidence_lo: pd.Series = None, confidence_hi: pd.Series = None):

        assert len(series) >= 3, 'Series must have at least three values.'  # cannot create a timeseries with n<3 -> can add less than 3 elements with add function
        assert isinstance(series.index, pd.DatetimeIndex), 'Series must be indexed with a DatetimeIndex.'
        assert np.issubdtype(series.dtype, np.number), 'Series must contain numerical values.'

        self._series: pd.Series = series.sort_index()  # Sort by time
        self._freq: str = self._series.index.inferred_freq  # Infer frequency

        # TODO: optionally fill holes (including missing dates) - for now we assume no missing dates
        assert self._freq is not None, 'Could not infer frequency. Are some dates missing? Is Series too short (n=2)?'

        # TODO: are there some pandas Series where the line below causes issues?
        self._series.index.freq = self._freq  # Set the inferred frequency in the Pandas series

        # Handle confidence intervals:
        self._confidence_lo = None
        self._confidence_hi = None
        if confidence_lo is not None:
            self._confidence_lo = confidence_lo.sort_index()
            assert len(self._confidence_lo) == len(self._series), 'Lower confidence interval must have same size as ' \
                                                                  'the main time series.'
            assert (self._confidence_lo.index == self._series.index).all(), 'Lower confidence interval and main ' \
                                                                            'series must have the same time index.'
        if confidence_hi is not None:
            self._confidence_hi = confidence_hi.sort_index()
            assert len(self._confidence_hi) == len(self._series), 'Upper confidence interval must have same size as ' \
                                                                  'the main time series.'
            assert (self._confidence_hi.index == self._series.index).all(), 'Upper confidence interval and main ' \
                                                                            'series must have the same time index.'

    def pd_series(self) -> pd.Series:
        """
        Returns the underlying Pandas Series of this TimeSeries.

        :return: A Pandas Series.
        """
        return self._series.copy()

    def conf_lo_pd_series(self) -> Optional[pd.Series]:
        """
        Returns the underlying Pandas Series of the lower confidence interval if it exists.

        :return: A Pandas Series for the lower confidence interval.
        """
        return self._confidence_lo.copy() if self._confidence_lo is not None else None

    def conf_hi_pd_series(self) -> Optional[pd.Series]:
        """
        Returns the underlying Pandas Series of the upper confidence interval if it exists.

        :return: A Pandas Series for the upper confidence interval.
        """
        return self._confidence_hi.copy() if self._confidence_hi is not None else None

    def start_time(self) -> pd.Timestamp:
        """
        Returns the start time of the time index.

        :return: A timestamp containing the first time of the TimeSeries.
        """
        return self._series.index[0]

    def end_time(self) -> pd.Timestamp:
        """
        Returns the end time of the time index.

        :return: A timestamp containing the last time of the TimeSeries.
        """
        return self._series.index[-1]

    def values(self) -> np.ndarray:
        """
        Returns the values of the TimeSeries.

        :return: A numpy array containing the values of the TimeSeries.
        """
        return self._series.values

    def time_index(self) -> pd.DatetimeIndex:
        """
        Returns the index of the TimeSeries.

        :return: A DatetimeIndex containing the index of the TimeSeries.
        """
        return self._series.index

    def freq(self) -> pd.DateOffset:
        """
        Returns the frequency of the TimeSeries.

        :return: A DateOffset with the frequency.
        """
        return to_offset(self._freq)

    def freq_str(self) -> str:
        """
        Returns the frequency of the TimeSeries.

        :return: A string with the frequency.
        """
        return self._freq

    def duration(self) -> pd.Timedelta:
        """
        Returns the duration of the TimeSeries.

        :return: A Timedelta of the duration of the TimeSeries.
        """
        return self._series.index[-1] - self._series.index[0]

    def copy(self, deep: bool = True):
        """
        Make a copy of this object time series
        :param deep: Make a deep copy. If False, the Series will be the same
        :return: A copy of the TimeSeries
        """
        if deep:
            return TimeSeries(self.pd_series(), self.conf_lo_pd_series(), self.conf_hi_pd_series())
        else:
            return TimeSeries(self._series, self._confidence_lo, self._confidence_hi)

    def _raise_if_not_within(self, ts: pd.Timestamp):

        if (ts < self.start_time()) or (ts > self.end_time()):
            raise ValueError('Timestamp must be between {} and {}'.format(self.start_time(), self.end_time()))

    def split_after(self, ts: pd.Timestamp) -> Tuple['TimeSeries', 'TimeSeries']:
        """
        Splits the TimeSeries in two, around a provided timestamp [ts].
        
        The timestamp may not be in the TimeSeries. If it is, the timestamp will be included in the
        first of the two TimeSeries, and not in the second.
        
        :param ts: The timestamp that indicates the splitting time.
        :return: A tuple (s1, s2) of TimeSeries with indices smaller or equal to [ts]
                 and greater than [ts] respectively.
        """

        self._raise_if_not_within(ts)

        ts = self.time_index()[self.time_index() <= ts][-1]  # closest index before ts (new ts)

        start_second_series: pd.Timestamp = ts + self.freq()  # second series does not include ts
        return self.slice(self.start_time(), ts), self.slice(start_second_series, self.end_time())

    def split_before(self, ts: pd.Timestamp) -> Tuple['TimeSeries', 'TimeSeries']:
        """
        Splits a TimeSeries in two, around a provided timestamp [ts].

        The timestamp may not be in the TimeSeries. If it is, the timestamp will be included in the
        second of the two TimeSeries, and not in the first.

        :param ts: The timestamp that indicates the splitting time.
        :return: A tuple (s1, s2) of TimeSeries with indices smaller than [ts]
                 and greater or equal to [ts] respectively.
        """
        self._raise_if_not_within(ts)

        ts = self.time_index()[self.time_index() >= ts][0]  # closest index after ts (new ts)

        end_first_series: pd.Timestamp = ts - self.freq()  # second series does not include ts
        return self.slice(self.start_time(), end_first_series), self.slice(ts, self.end_time())

    def drop_after(self, ts: pd.Timestamp) -> 'TimeSeries':
        """
        Drops everything after the provided timestamp [ts], included.

        The timestamp may not be in the TimeSeries. If it is, the timestamp will be dropped.

        :param ts: The timestamp that indicates cut-off time.
        :return: A new TimeSeries, with indices smaller than [ts].
        """
        self._raise_if_not_within(ts)

        ts = self.time_index()[self.time_index() >= ts][0]  # closest index after ts (new ts)

        end_series: pd.Timestamp = ts - self.freq()  # new series does not include ts
        return self.slice(self.start_time(), end_series)

    def drop_before(self, ts: pd.Timestamp) -> 'TimeSeries':
        """
        Drops everything before the provided timestamp [ts], included.

        The timestamp may not be in the TimeSeries. If it is, the timestamp will be dropped.

        :param ts: The timestamp that indicates cut-off time.
        :return: A new TimeSeries, with indices greater than [ts].
        """
        self._raise_if_not_within(ts)

        ts = self.time_index()[self.time_index() <= ts][-1]  # closest index before ts (new ts)

        start_series: pd.Timestamp = ts + self.freq()  # new series does not include ts
        return self.slice(start_series, self.end_time())

    def slice(self, start_ts: pd.Timestamp, end_ts: pd.Timestamp) -> 'TimeSeries':
        """
        Returns a new TimeSeries, starting later than [start_ts] and ending before [end_ts], inclusive on both ends.

        The timestamps may not be in the time series. TODO: should new timestamps be added? Think not

        :param start_ts: The timestamp that indicates the left cut-off.
        :param end_ts: The timestamp that indicates the right cut-off.
        :return: A new TimeSeries, which indices greater or equal than [start_ts] and smaller or equal than [end_ts].
        """

        assert end_ts > start_ts, 'End timestamp must be strictly after start timestamp when slicing.'
        assert end_ts >= self.start_time(), 'End timestamp must be after the start of the time series when slicing.'
        assert start_ts <= self.end_time(), 'Start timestamp must be after the end of the time series when slicing.'

        def _slice_not_none(s: Optional[pd.Series]) -> Optional[pd.Series]:
            if s is not None:
                s_a = s[s.index >= start_ts]
                return s_a[s_a.index <= end_ts]
            return None

        return TimeSeries(_slice_not_none(self._series),
                          _slice_not_none(self._confidence_lo),
                          _slice_not_none(self._confidence_hi))

    def slice_n_points_after(self, start_ts: pd.Timestamp, n: int) -> 'TimeSeries':
        """
        Returns a new TimeSeries, starting later than [start_ts] (included) and having (at most) [n] points.

        The timestamp may not be in the time series. If it is, it will be included in the new TimeSeries.

        :param start_ts: The timestamp that indicates the splitting time.
        :param n: The maximal length of the new TimeSeries.
        :return: A new TimeSeries, with length at most [n] and indices greater or equal than [start_ts].
        """

        assert n >= 0, 'n should be a positive integer.'  # TODO: logically raise if n<3, cf. init

        self._raise_if_not_within(start_ts)

        start_ts = self.time_index()[self.time_index() >= start_ts][0]  # closest index after start_ts (new start_ts)

        end_ts: pd.Timestamp = start_ts + (n - 1) * self.freq()  # (n-1) because slice() is inclusive on both sides
        return self.slice(start_ts, end_ts)

    def slice_n_points_before(self, end_ts: pd.Timestamp, n: int) -> 'TimeSeries':
        """
        Returns a new TimeSeries, ending before [end_ts] (included) and having (at most) [n] points.

        The timestamp may not be in the TimeSeries. If it is, it will be included in the new TimeSeries.

        :param end_ts: The timestamp that indicates the splitting time.
        :param n: The maximal length of the new time series.
        :return: A new TimeSeries, with length at most [n] and indices smaller or equal than [end_ts].
        """

        assert n >= 0, 'n should be a positive integer.'

        self._raise_if_not_within(end_ts)

        end_ts = self.time_index()[self.time_index() <= end_ts][-1]

        start_ts: pd.Timestamp = end_ts - (n - 1) * self.freq()  # (n-1) because slice() is inclusive on both sides
        return self.slice(start_ts, end_ts)

    def intersect(self, other: 'TimeSeries') -> 'TimeSeries':
        """
        Returns a slice containing the intersection of this TimeSeries and the one provided in argument.

        TODO: This function does not really behave as described. Not really an intersect

        :param other: A second TimeSeries.
        :return: A new TimeSeries, with values of this TimeSeries and indices the intersection of both
                TimeSeries' indices.
        """

        def _intersect_not_none(s: Optional[pd.Series]) -> Optional[pd.Series]:
            if s is not None:
                new_index = self.time_index().intersection(other.time_index())

                assert len(s) > 2, 'The two series do not have enough common times.'

                return s[new_index]
            return None

        return TimeSeries(_intersect_not_none(self._series),
                          _intersect_not_none(self._confidence_lo),
                          _intersect_not_none(self._confidence_hi))

    # TODO: other rescale? such as giving a ratio, or a specific position? Can be the same function
    def rescale_with_value(self, value_at_first_step: float) -> 'TimeSeries':
        """
        Returns a new TimeSeries, which is a multiple of this TimeSeries such that
        the first value is [value_at_first_step].
        Numerical imprecisions appear with [value_at_first_step] > 1e+24

        :param value_at_first_step: The new value for the first entry of the TimeSeries.
        :return: A new TimeSeries, whose first value was changed to [value_at_first_step] and whose others values
                have been scaled accordingly.
        """

        assert self.values()[0] != 0, 'Cannot rescale with first value 0.'

        coef = value_at_first_step / self.values()[0]  # TODO: should the new TimeSeries have the same dtype?
        new_series = coef * self._series
        new_conf_lo = self._op_or_none(self._confidence_lo, lambda s: s * coef)
        new_conf_hi = self._op_or_none(self._confidence_hi, lambda s: s * coef)
        return TimeSeries(new_series, new_conf_lo, new_conf_hi)

    def shift(self, n: int) -> 'TimeSeries':
        """
        Shifts the time axis of this TimeSeries by [n] time steps.

        If n > 0, shifts in the future. If n < 0, shifts in the past.

        For example, with n=2 and freq='M', March 2013 becomes May 2013. With n=-2, March 2013 becomes Jan 2013.

        :param n: The signed number of time steps to shift by.
        :return: A new TimeSeries, with a shifted index.
        """
        # TODO: no error raised if freq is different than day and overflow happens, so raise it ourselves
        try:
            self.time_index()[-1] + n * self.freq()
        except pd.errors.OutOfBoundsDatetime:
            raise OverflowError("the add operation between {} and {} will overflow".format(n * self.freq(),
                                                                                           self.time_index()[-1]))
        new_time_index = self._series.index.map(lambda ts: ts + n * self.freq())
        new_series = self._series.copy()
        new_series.index = new_time_index
        new_conf_lo = None
        new_conf_hi = None
        if self._confidence_lo is not None:
            new_conf_lo = self._confidence_lo.copy()
            new_conf_lo.index = new_time_index
        if self._confidence_hi is not None:
            new_conf_hi = self._confidence_hi.copy()
            new_conf_hi.index = new_time_index
        return TimeSeries(new_series, new_conf_lo, new_conf_hi)

    @staticmethod
    def from_dataframe(df: pd.DataFrame, time_col: str, value_col: str,
                       conf_lo_col: str = None, conf_hi_col: str = None) -> 'TimeSeries':
        """
        Returns a TimeSeries built from some DataFrame columns (ignoring DataFrame index)

        :param df: The DataFrame
        :param time_col: The time column name (mandatory).
        :param value_col: The value column name (mandatory).
        :param conf_lo_col: The lower confidence interval column name (optional).
        :param conf_hi_col: The upper confidence interval column name (optional).
        :return: A TimeSeries constructed from the inputs.
        """
        # TODO: return a list, object, ... of TimeSeries if there are multiple features (value-col). To decide

        times: pd.Series = pd.to_datetime(df[time_col], errors='raise')
        series: pd.Series = pd.Series(df[value_col].values, index=times)

        conf_lo = pd.Series(df[conf_lo_col], index=times) if conf_lo_col is not None else None
        conf_hi = pd.Series(df[conf_hi_col], index=times) if conf_hi_col is not None else None

        return TimeSeries(series, conf_lo, conf_hi)

    @staticmethod
    def from_times_and_values(times: pd.DatetimeIndex,
                              values: np.ndarray,
                              confidence_lo: np.ndarray = None,
                              confidence_hi: np.ndarray = None) -> 'TimeSeries':
        """
        Returns a TimeSeries built from an index and values.

        :param times: A DateTimeIndex for the TimeSeries.
        :param values: An array of values for the TimeSeries.
        :param confidence_lo: The lower confidence interval values (optional).
        :param confidence_hi: The higher confidence interval values (optional).
        :return: A TimeSeries constructed from the inputs.
        """

        series = pd.Series(values, index=times)
        series_lo = pd.Series(confidence_lo, index=times) if confidence_lo is not None else None
        series_hi = pd.Series(confidence_hi, index=times) if confidence_hi is not None else None

        return TimeSeries(series, series_lo, series_hi)

    def plot(self, *args, plot_ci=True, **kwargs):
        """
        Currently this is just a wrapper around pd.Series.plot()
        """
        # temporary work-around for the pandas.plot issue
        # errors = self._combine_or_none(self._confidence_lo, self._confidence_hi,
        #                                lambda x, y: np.vstack([x.values, y.values]))
        # self._series.plot(yerr=errors, *args, **kwargs)
        plt.plot(self.time_index(), self.values(), *args, **kwargs)
        x_label = self.time_index().name
        if x_label is not None and len(x_label) > 0:
            plt.xlabel(x_label)
        # TODO: use pandas plot in the future
        if plot_ci and self._confidence_lo is not None and self._confidence_hi is not None:
            plt.fill_between(self.time_index(), self._confidence_lo.values, self._confidence_hi.values, alpha=0.5)

    """
    Some useful methods for TimeSeries combination:
    """

    def has_same_time_as(self, other: 'TimeSeries') -> bool:
        """
        Checks whether this TimeSeries and another one have the same index.

        :param other: A second TimeSeries.
        :return: A boolean. True if both TimeSeries have the same index, False otherwise.
        """

        if self.__len__() != len(other):
            return False
        return (other.time_index() == self.time_index()).all()

    # TODO: is union function useful too?
    # TODO: should append only at the end of the series? or can we create holes and "interpolate" their values?
    def append(self, other: 'TimeSeries') -> 'TimeSeries':
        """
        Appends another TimeSeries to this TimeSeries.

        :param other: A second TimeSeries.
        :return: A new TimeSeries, obtained by appending the second TimeSeries to the first.
        """

        assert other.start_time() == self.end_time() + self.freq(), 'Appended TimeSeries must start one time step ' \
                                                                    'after current one.'
        # TODO additional check?
        assert other.freq() == self.freq(), 'Appended TimeSeries must have the same frequency as the current one'

        series = self._series.append(other.pd_series())
        conf_lo = None
        conf_hi = None
        if self._confidence_lo is not None and other.conf_lo_pd_series() is not None:
            conf_lo = self._confidence_lo.append(other.conf_lo_pd_series())
        if self._confidence_hi is not None and other.conf_hi_pd_series() is not None:
            conf_hi = self._confidence_hi.append(other.conf_hi_pd_series())
        return TimeSeries(series, conf_lo, conf_hi)

    def append_values(self, values: np.ndarray, index: pd.DatetimeIndex = None,
            conf_lo: np.ndarray = None, conf_hi: np.ndarray = None) -> 'TimeSeries':
        """
        Appends values to current TimeSeries, to the given indices.

        If no index is provided, assumes that it follows the original data.
        Does not add new confidence values if there were none first.
        Does not update value if already existing indices are provided.

        :param values: An array with the values to append.
        :param index: A DateTimeIndex for each value (optional).
        :param conf_lo: The lower confidence interval values (optional).
        :param conf_hi: The higher confidence interval values (optional).
        :return: A new TimeSeries with the new values appended
        """
        if len(values) < 1:
            return self
        if isinstance(values, list):
            values = np.array(values)
        if index is None:
            index = pd.DatetimeIndex([self.end_time() + i * self.freq() for i in range(1, 1+len(values))])
        assert isinstance(index, pd.DatetimeIndex), 'values must be indexed with a DatetimeIndex.'
        assert len(index) == len(values)
        assert self.time_index().intersection(index).empty, "cannot add already present time index"
        new_indices = index.argsort()
        index = index[new_indices]
        # TODO do we really want that?
        assert index[0] == self.end_time() + self.freq(), 'Appended index must start one time step ' \
                                                          'after current one.'
        if len(index) > 2:
            assert index.inferred_freq == self.freq_str(), 'Appended index must have ' \
                                                           'the same frequency as the current one'
        elif len(index) == 2:
            assert index[-1] == index[0] + self.freq(), 'Appended index must have ' \
                                                        'the same frequency as the current one'
        values = values[new_indices]
        new_series = pd.Series(values, index=index)
        series = self._series.append(new_series)
        if conf_lo is not None and self._confidence_lo is not None:
            assert len(index) == len(conf_lo)
            conf_lo = conf_lo[new_indices]
            conf_lo = self._confidence_lo.append(pd.Series(conf_lo, index=index))
        if conf_hi is not None and self._confidence_hi is not None:
            assert len(index) == len(conf_hi)
            conf_hi = conf_hi[new_indices]
            conf_hi = self._confidence_hi.append(pd.Series(conf_hi, index=index))

        return TimeSeries(series, conf_lo, conf_hi)

    def update(self, index: pd.DatetimeIndex, values: np.ndarray = None,
               conf_lo: np.ndarray = None, conf_hi: np.ndarray = None, inplace: bool = True) -> 'TimeSeries':
        """
        Updates the Series with the new values provided.
        If indices are not in original TimeSeries, they will be discarded.
        At least one parameter other than index must be filled.
        Use np.nan to ignore a specific index in a series.

        It will raise an error if try to update a missing CI series

        :param index: A DateTimeIndex containing the indices to replace.
        :param values: An array containing the values to replace (optional).
        :param conf_lo: The lower confidence interval values to change (optional).
        :param conf_hi: The higher confidence interval values (optional).
        :param inplace: If True, do operation inplace and return None, defaults to True.
        :return: A TimeSeries with values updated
        """
        assert not (values is None and conf_lo is None and conf_hi is None), "At least one parameter must be filled " \
                                                                             "other than index"
        assert True if values is None else len(values) == len(index), \
            "The number of values must correspond to the number of indices: {} != {}".format(len(values), len(index))
        assert True if conf_lo is None else len(conf_lo) == len(index), \
            "The number of values must correspond to the number of indices: {} != {}".format(len(conf_lo), len(index))
        assert True if conf_hi is None else len(conf_hi) == len(index), \
            "The number of values must correspond to the number of indices: {} != {}".format(len(conf_hi), len(index))
        ignored_indices = [index.get_loc(ind) for ind in (set(index)-set(self.time_index()))]
        index = index.delete(ignored_indices)
        series = values if values is None else pd.Series(np.delete(values, ignored_indices), index=index)
        conf_lo = conf_lo if conf_lo is None else pd.Series(np.delete(conf_lo, ignored_indices), index=index)
        conf_hi = conf_hi if conf_hi is None else pd.Series(np.delete(conf_hi, ignored_indices), index=index)
        assert len(index) > 0, "must give at least one correct index"
        if inplace:
            if series is not None:
                self._series.update(series)
            if conf_lo is not None:
                self._confidence_lo.update(conf_lo)
            if conf_hi is not None:
                self._confidence_hi.update(conf_hi)
            return None
        else:
            new_series = self.pd_series()
            new_lo = self.conf_lo_pd_series()
            new_hi = self.conf_hi_pd_series()
            if series is not None:
                new_series.update(series)
            if conf_lo is not None:
                new_lo.update(conf_lo)
            if conf_hi is not None:
                new_hi.update(conf_hi)
            return TimeSeries(new_series, new_lo, new_hi)

    def drop_values(self, index: pd.DatetimeIndex, inplace: bool = True, **kwargs):
        """
        Remove elements of all series with specified indices.

        :param index: The indices to be dropped
        :param kwargs: Option to pass to pd.Series drop method
        :param inplace: If True, do operation inplace and return None, defaults to True.
        :return: A TimeSeries with values dropped
        """
        series = self._series.drop(index=index, inplace=inplace, **kwargs)
        conf_lo = self._op_or_none(self._confidence_lo, lambda s: s.drop(index, inplace=inplace, **kwargs))
        conf_hi = self._op_or_none(self._confidence_hi, lambda s: s.drop(index, inplace=inplace, **kwargs))
        if inplace:
            return None
        return TimeSeries(series, conf_lo, conf_hi)

    @staticmethod
    def _combine_or_none(series_a: Optional[pd.Series],
                         series_b: Optional[pd.Series],
                         combine_fn: Callable[[pd.Series, pd.Series], Any]):
        """
        Combines two Pandas Series [series_a] and [series_b] using [combine_fn] if neither is None.

        :param series_a: A Pandas Series.
        :param series_b: A Pandas Series.
        :param combine_fn: An operation with input two Pandas Series and output one Pandas Series.
        :return: A new Pandas Series, the result of [combine_fn], or None.
        """

        if series_a is not None and series_b is not None:
            return combine_fn(series_a, series_b)
        return None

    @staticmethod
    def _op_or_none(series: Optional[pd.Series], op: Callable[[pd.Series], Any]):
        return op(series) if series is not None else None

    def _combine_from_pd_ops(self, other: 'TimeSeries',
                             combine_fn: Callable[[pd.Series, pd.Series], pd.Series]) -> 'TimeSeries':
        """
        Combines this TimeSeries with another one, using the [combine_fn] on the underlying Pandas Series.

        :param other: A second TimeSeries.
        :param combine_fn: An operation with input two Pandas Series and output one Pandas Series.
        :return: A new TimeSeries, with underlying Pandas Series the series obtained with [combine_fn].
        """

        assert self.has_same_time_as(other), 'The two TimeSeries must have the same time index.'

        series = combine_fn(self._series, other.pd_series())
        conf_lo = self._combine_or_none(self._confidence_lo, other.conf_lo_pd_series(), combine_fn)
        conf_hi = self._combine_or_none(self._confidence_hi, other.conf_hi_pd_series(), combine_fn)
        return TimeSeries(series, conf_lo, conf_hi)

    """
    Definition of some useful statistical methods.
    
    These methods rely on the Pandas implementation.
    """

    def mean(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.mean(axis, skipna, level, numeric_only, **kwargs)

    def var(self, axis=None, skipna=None, level=None, ddof=1, numeric_only=None, **kwargs) -> float:
        return self._series.var(axis, skipna, level, ddof, numeric_only, **kwargs)

    def std(self, axis=None, skipna=None, level=None, ddof=1, numeric_only=None, **kwargs) -> float:
        return self._series.std(axis, skipna, level, ddof, numeric_only, **kwargs)

    def skew(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.skew(axis, skipna, level, numeric_only, **kwargs)

    def kurtosis(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.kurtosis(axis, skipna, level, numeric_only, **kwargs)

    def min(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.min(axis, skipna, level, numeric_only, **kwargs)

    def max(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.max(axis, skipna, level, numeric_only, **kwargs)

    def sum(self, axis=None, skipna=None, level=None, numeric_only=None, min_count=0, **kwargs) -> float:
        return self._series.sum(axis, skipna, level, numeric_only, min_count, **kwargs)

    def median(self, axis=None, skipna=None, level=None, numeric_only=None, **kwargs) -> float:
        return self._series.median(axis, skipna, level, numeric_only, **kwargs)

    def autocorr(self, lag=1) -> float:
        return self._series.autocorr(lag)

    def describe(self, percentiles=None, include=None, exclude=None) -> pd.Series:
        return self._series.describe(percentiles, include, exclude)

    """
    Definition of some dunder methods
    """

    def __eq__(self, other):
        if isinstance(other, TimeSeries):
            if not self._series.equals(other.pd_series()):
                return False
            for other_ci, self_ci in zip([other.conf_lo_pd_series(), other.conf_hi_pd_series()],
                                [self._confidence_lo, self._confidence_hi]):
                if (other_ci is None) ^ (self_ci is None):
                    # only one is None
                    return False
                if self._combine_or_none(self_ci, other_ci, lambda s1, s2: s1.equals(s2)) is False:
                    # Note: we check for "False" explicitly, because None is OK..
                    return False
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self._series)

    def __add__(self, other):
        if isinstance(other, (int, float)):
            new_series = self._series + other
            conf_lo = self._op_or_none(self._confidence_lo, lambda s: s + other)
            conf_hi = self._op_or_none(self._confidence_hi, lambda s: s + other)
            return TimeSeries(new_series, conf_lo, conf_hi)
        elif isinstance(other, TimeSeries):
            return self._combine_from_pd_ops(other, lambda s1, s2: s1 + s2)
        else:
            raise TypeError('unsupported operand type(s) for + or add(): \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            new_series = self._series - other
            conf_lo = self._op_or_none(self._confidence_lo, lambda s: s - other)
            conf_hi = self._op_or_none(self._confidence_hi, lambda s: s - other)
            return TimeSeries(new_series, conf_lo, conf_hi)
        elif isinstance(other, TimeSeries):
            return self._combine_from_pd_ops(other, lambda s1, s2: s1 - s2)
        else:
            raise TypeError('unsupported operand type(s) for - or sub(): \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))

    def __rsub__(self, other):
        return other + (-self)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            new_series = self._series * other
            conf_lo = self._op_or_none(self._confidence_lo, lambda s: s * other)
            conf_hi = self._op_or_none(self._confidence_hi, lambda s: s * other)
            return TimeSeries(new_series, conf_lo, conf_hi)
        elif isinstance(other, TimeSeries):
            return self._combine_from_pd_ops(other, lambda s1, s2: s1 * s2)
        else:
            raise TypeError('unsupported operand type(s) for * or mul(): \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))

    def __rmul__(self, other):
        return self * other

    def __pow__(self, n):
        if isinstance(n, (int, float)):
            if n < 0:
                assert all(self.values() != 0), 'Cannot divide by a TimeSeries with a value 0.'

            new_series = self._series ** float(n)
            conf_lo = self._op_or_none(self._confidence_lo, lambda s: s ** float(n))
            conf_hi = self._op_or_none(self._confidence_hi, lambda s: s ** float(n))
            return TimeSeries(new_series, conf_lo, conf_hi)
        else:
            raise TypeError('unsupported operand type(s) for ** or pow(): \'{}\' and \'{}\'.' \
                            .format(type(self).__name__, type(n).__name__))

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            assert other != 0, 'Cannot divide by 0.'

            new_series = self._series / other
            conf_lo = self._op_or_none(self._confidence_lo, lambda s: s / other)
            conf_hi = self._op_or_none(self._confidence_hi, lambda s: s / other)
            return TimeSeries(new_series, conf_lo, conf_hi)

        elif isinstance(other, TimeSeries):
            assert all(other.values() != 0), 'Cannot divide by a TimeSeries with a value 0.'

            return self._combine_from_pd_ops(other, lambda s1, s2: s1 / s2)
        else:
            raise TypeError('unsupported operand type(s) for / or truediv(): \'{}\' and \'{}\'.' \
                            .format(type(self).__name__, type(other).__name__))

    def __rtruediv__(self, n):
        return n * (self ** (-1))

    def __abs__(self):
        series = abs(self._series)
        conf_lo = self._op_or_none(self._confidence_lo, lambda s: abs(s))
        conf_hi = self._op_or_none(self._confidence_hi, lambda s: abs(s))
        return TimeSeries(series, conf_lo, conf_hi)

    def __neg__(self):
        series = -self._series
        conf_lo = self._op_or_none(self._confidence_lo, lambda s: -s)
        conf_hi = self._op_or_none(self._confidence_hi, lambda s: -s)
        return TimeSeries(series, conf_lo, conf_hi)

    def __contains__(self, item):
        if isinstance(item, pd.Timestamp):
            return item in self._series.index
        return False

    def __round__(self, n=None):
        series = self._series.round(n)
        confidence_lo = self._op_or_none(self._confidence_lo, lambda s: s.round(n))
        confidence_hi = self._op_or_none(self._confidence_hi, lambda s: s.round(n))
        return TimeSeries(series, confidence_lo, confidence_hi)

    # TODO: Ignoring confidence series for now
    def __lt__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            series = self._series < other
        elif isinstance(other, TimeSeries):
            series = self._series < other.pd_series()
        else:
            raise TypeError('unsupported operand type(s) for < : \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))
        return series  # TODO should we return only the ndarray, the pd series, or our timeseries?

    def __gt__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            series = self._series > other
        elif isinstance(other, TimeSeries):
            series = self._series > other.pd_series()
        else:
            raise TypeError('unsupported operand type(s) for > : \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))
        return series

    def __le__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            series = self._series <= other
        elif isinstance(other, TimeSeries):
            series = self._series <= other.pd_series()
        else:
            raise TypeError('unsupported operand type(s) for <= : \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))
        return series

    def __ge__(self, other):
        if isinstance(other, (int, float, np.ndarray)):
            series = self._series >= other
        elif isinstance(other, TimeSeries):
            series = self._series >= other.pd_series()
        else:
            raise TypeError('unsupported operand type(s) for >= : \'{}\' and \'{}\'.'
                            .format(type(self).__name__, type(other).__name__))
        return series

    def __str__(self):
        df = pd.DataFrame({'value': self._series})
        if self._confidence_lo is not None:
            df['conf_low'] = self._confidence_lo
        if self._confidence_hi is not None:
            df['conf_high'] = self._confidence_hi
        return str(df) + '\nFreq: {}'.format(self.freq_str())

    def __repr__(self):
        return self.__str__()

    def __copy__(self, deep: bool = True):
        return self.copy(deep=deep)

    def __deepcopy__(self, memodict={}):
        return self.copy(deep=True)

    def __getitem__(self, item):
        # return only main series if nb of values < 3
        if isinstance(item, (int, pd.Timestamp)):
            return self._series[[item]]
        elif isinstance(item, (pd.DatetimeIndex, slice, list, np.ndarray)):
            if isinstance(item, slice):
                # if create a slice with timestamp, convert to indices
                if item.start.__class__ == pd.Timestamp or item.stop.__class__ == pd.Timestamp:
                    istart = None if item.start is None else self.time_index().get_loc(item.start)
                    istop = None if item.stop is None else self.time_index().get_loc(item.stop)
                    item = slice(istart, istop, item.step)
                # cannot reverse order
                if item.indices(len(self))[-1] == -1:
                    raise IndexError("Cannot have a backward TimeSeries")
            # Verify that values in item are really in index to avoid the creation of NaN values
            if isinstance(item, (np.ndarray, pd.DatetimeIndex)):
                check = np.array([elem in self.time_index() for elem in item])
                if not np.all(check):
                    raise IndexError("None of {} in the index".format(item[~check]))
            try:
                return TimeSeries(self._series[item],
                                  self._op_or_none(self._confidence_lo, lambda s: s[item]),
                                  self._op_or_none(self._confidence_hi, lambda s: s[item]))
            except AssertionError:
                # return only main series if nb of values < 3
                return self._series[item]
        else:
            raise IndexError("Input {} of class {} is not a possible key.\n"\
                             "Please use integers, pd.DateTimeIndex, arrays or slice".format(item, item.__class__))



from .autoregressive_model import AutoRegressiveModel
from statsmodels.tsa.arima_model import ARMA, ARIMA
from pmdarima import auto_arima
from ..timeseries import TimeSeries


class Arima(AutoRegressiveModel):
    """
    Implementation of an ARIMA model.

    Currently a wrapper around the statsmodel implementation.

    :param p: An integer representing the lag order.
    :param d: An integer for the order of differentiation.
    :param q: An interger for the size of the moving average window.
    """

    def __init__(self, p: int = 12, d: int = 1, q: int = 0):
        super().__init__()
        self.p = p
        self.d = d
        self.q = q
        self.model = None

    def __str__(self):
        return 'ARIMA({},{},{})'.format(self.p, self.d, self.q)

    def fit(self, series: TimeSeries):
        super().fit(series)

        m = ARIMA(series.values(),
                  order=(self.p, self.d, self.q)) if self.d > 0 else ARMA(series.values(), order=(self.p, self.q))
        self.model = m.fit(disp=0)

    def predict(self, n):
        super().predict(n)
        forecast = self.model.forecast(steps=n)[0]
        return self._build_forecast_series(forecast)


class AutoArima(AutoRegressiveModel):

    def __init__(self, start_p=1, max_p=12, start_q=0, max_q=12, max_P=2, max_Q=2, start_P=1, start_Q=1,
                 start_d=0, max_d=2, max_D=1, max_order=30, seasonal=True, stepwise=True, approximation=False,
                 error_action='ignore', trace=False, suppress_warnings=True):

        super().__init__()

        self.start_p = start_p
        self.max_p = max_p
        self.start_q = start_q
        self.max_q = max_q
        self.max_P = max_P
        self.max_Q = max_Q
        self.start_P = start_P
        self.start_Q = start_Q
        self.start_d = start_d
        self.max_d = max_d
        self.max_D = max_D
        self.max_order = max_order
        self.seasonal = seasonal
        self.stepwise = stepwise
        self.approximation = approximation
        self.error_action = error_action
        self.trace = trace
        self.suppress_warnings = suppress_warnings
        self.model = None

    def __str__(self):
        return 'auto-ARIMA'

    def fit(self, series: TimeSeries):
        super().fit(series)
        self.model = auto_arima(series.values(),
                                start_p=self.start_p,
                                max_p=self.max_p,
                                start_q=self.start_q,
                                max_q=self.max_q,
                                max_P=self.max_P,
                                max_Q=self.max_Q,
                                start_P=self.start_P,
                                start_Q=self.start_Q,
                                start_d=self.start_d,
                                max_d=self.max_d,
                                max_D=self.max_D,
                                max_order=self.max_order,
                                seasonal=self.seasonal,
                                stepwise=self.stepwise,
                                approximation=self.approximation,
                                error_action=self.error_action,
                                trace=self.trace,
                                suppress_warnings=self.suppress_warnings)

    def predict(self, n):
        super().predict(n)
        forecast = self.model.predict(n_periods=n)
        return self._build_forecast_series(forecast)



from ..timeseries import TimeSeries
from typing import List


def retain_period_common_to_all(series: List[TimeSeries]) -> List[TimeSeries]:
    """
    Trims all series in the provided list, if necessary, so that the return time series have
    the same time index (corresponding to largest duration common to all series).

    Raises an error if no such time index exists.
    :param series:
    :return:
    """

    last_first = max(map(lambda s: s.start_time(), series))
    first_last = min(map(lambda s: s.end_time(), series))

    if last_first >= first_last:
        raise ValueError('The provided time series must have nonzero overlap')

    return list(map(lambda s: s.slice(last_first, first_last), series))



import pandas as pd
from u8timeseries.models.baselines import KthValueAgoBaseline
from u8timeseries.timeseries import TimeSeries

my_df = pd.DataFrame({
  'timestep': [1, 2, 3],
  'values': [4, 5, 6]
})

series = TimeSeries.from_dataframe(my_df, 'timestep', 'values')
series.plot()

my_model = KthValueAgoBaseline(K=1)
my_model.fit(series)

# n step ahead prediction:
predictions = my_model.predict(n=6)
print(predictions)



import pandas as pd
import numpy as np
from IPython import get_ipython
from tqdm import tqdm, tqdm_notebook
from u8timeseries.timeseries import TimeSeries
from u8timeseries.models.autoregressive_model import AutoRegressiveModel
from u8timeseries.models.regressive_model import RegressiveModel
from typing import List


def _build_iterator(iterable, verbose):
    def _isnotebook():
        try:
            shell = get_ipython().__class__.__name__
            if shell == 'ZMQInteractiveShell':
                return True  # Jupyter notebook or qtconsole
            elif shell == 'TerminalInteractiveShell':
                return False  # Terminal running IPython
            else:
                return False  # Other type (?)
        except NameError:
            return False  # Probably standard Python interpreter

    if verbose:
        if _isnotebook():
            iterator = tqdm_notebook(iterable)
        else:
            iterator = tqdm(iterable)
    else:
        iterator = iterable
    return iterator


def simulate_forecast_ar(series: 'TimeSeries',
                         model: 'AutoRegressiveModel',
                         start: 'pd.Timestamp',
                         fcast_horizon_n: int,
                         trim_to_series: bool = True,
                         verbose=False) -> 'TimeSeries':
    """
    Provides an environment for forecasting future values of the TimeSeries 'series`.

    This function predicts the `fcast_horizon_n` values for the TimeSeries `series` starting from the date `start`
    according to the auto-regressive model `model`.

    :param series: The TimeSeries to forecast.
    :param model: The AutoRegressiveModel to use.
    :param start: The first time at which a prediction is produced for a future time.
    :param fcast_horizon_n: The number of future values to predict.
    :param trim_to_series: Whether the predicted series has the end trimmed to match the end of the main series or not.
    :param verbose: Whether to print progress or not.
    :return: A TimeSeries containing the fore-casted values of `series` over the horizon with respect to the model \
    `model`.

    """
    assert start in series, 'The provided start timestamp is not in the time series.'
    assert start != series.end_time(), 'The provided start timestamp is the last timestamp of the time series'

    last_pred_time = series.time_index()[-fcast_horizon_n - 2] if trim_to_series else series.time_index()[-2]

    # build the prediction times in advance (to be able to use tqdm)
    pred_times = [start]
    while pred_times[-1] <= last_pred_time:
        pred_times.append(pred_times[-1] + series.freq())

    # what we'll return
    values = []
    times = []

    iterator = _build_iterator(pred_times, verbose)

    for pred_time in iterator:
        if not verbose:
            print('.', end='')
        train = series.drop_after(pred_time)  # build the training series

        model.fit(train)
        pred = model.predict(fcast_horizon_n)
        values.append(pred.values()[-1])  # store the N-th point
        times.append(pred.end_time())  # store the N-th timestamp

    return TimeSeries.from_times_and_values(pd.DatetimeIndex(times), np.array(values))


def simulate_forecast_regr(feature_series: List[TimeSeries],
                           target_series: TimeSeries,
                           model: RegressiveModel,
                           start: pd.Timestamp,
                           fcast_horizon_n: int,
                           trim_to_series: bool = True,
                           verbose=False) -> TimeSeries:
    """
    Returns a TimeSeries containing the forecasts that would have been obtained from a given RegressiveModel,
    on a given forecast time horizon.

    .. todo: review and add to documentation.
    .. todo: optionally also return weights, when those are available in model
    .. todo: (getattr(model.model, 'coef_', None) is not None)

    :param feature_series: the feature time series of the regressive model
    :param target_series: the target time series of the regressive model (i.e., the series to predict)
    :param model: the RegressiveModel to use
    :param start: when the forecasts start (i.e., the first time at which a prediction is produced for a future time)
    :param fcast_horizon_n: the forecast horizon
    :param trim_to_series: whether the returned predicted series has the end trimmed to match the end of the main series
    :param verbose: whether to print progress
    :return:
    """
    assert all([s.has_same_time_as(target_series) for s in feature_series]), 'All provided time series must ' \
                                                                             'have the same time index'
    assert start in target_series, 'The provided start timestamp is not in the time series.'
    assert start != target_series.end_time(), 'The provided start timestamp is the last timestamp of the time series'

    last_pred_time = target_series.time_index()[-fcast_horizon_n - 2] if trim_to_series else target_series.time_index()[-2]

    # build the prediction times in advance (to be able to use tqdm)
    pred_times = [start]
    while pred_times[-1] <= last_pred_time:
        pred_times.append(pred_times[-1] + target_series.freq())

    # what we'll return
    values = []
    times = []

    iterator = _build_iterator(pred_times, verbose)

    for pred_time in iterator:
        if not verbose:
            print('.', end='')
        # build train/val series
        train_features = [s.drop_after(pred_time) for s in feature_series]
        train_target = target_series.drop_after(pred_time)
        val_features = [s.slice_n_points_after(pred_time + target_series.freq(), fcast_horizon_n)
                        for s in feature_series]

        model.fit(train_features, train_target)
        pred = model.predict(val_features)
        values.append(pred.values()[-1])  # store the N-th point
        times.append(pred.end_time())  # store the N-th timestamp

    return TimeSeries.from_times_and_values(pd.DatetimeIndex(times), np.array(values))


from .autoregressive_model import AutoRegressiveModel
from ..timeseries import TimeSeries
import statsmodels.tsa.holtwinters as hw


class ExponentialSmoothing(AutoRegressiveModel):
    """
    Implementation of a Simple Exponential Smoothing.

    Currently just a wrapper around the statsmodels holtwinter implementation.

    :param trend: A string for the type of trend to consider: either `additive` (default) or `multiplicative`.
    :param seasonal: A string for the type of seasonality to consider: either `additive` (default) or `multiplicative`.
    :param seasonal_periods: An integer, the order of seasonality to consider.
    """

    def __init__(self, trend: str = 'additive', seasonal: str = 'additive', seasonal_periods: int = 12):
        super().__init__()
        self.trend = trend
        self.seasonal = seasonal
        self.seasonal_periods = seasonal_periods
        self.model = None

    def __str__(self):
        return 'Exponential smoothing'

    def fit(self, series: TimeSeries):
        super().fit(series)
        self.model = hw.ExponentialSmoothing(series.values(),
                                             trend=self.trend,
                                             seasonal=self.seasonal,
                                             seasonal_periods=self.seasonal_periods).fit()

    def predict(self, n):
        super().predict(n)
        forecast = self.model.forecast(n)
        return self._build_forecast_series(forecast)



import pandas as pd
from u8timeseries.timeseries import TimeSeries
from u8timeseries.models.autoregressive_model import AutoRegressiveModel
from typing import Tuple, List, Callable, Any

""" This file contains some backtesting utilities, to evaluate user-defined functions (such as error functions)
    on some validation sets, sliding over time.
    The functionality is here for AutoRegressiveModel, not yet for regressive models.
    
    These functions will probably rarely be useful, compared to the ones in "forecasting_simulation.py"
"""


def get_train_val_series(series: TimeSeries, start: pd.Timestamp, nr_points_val: int,
                         nr_steps_iter: int = 1) -> List[Tuple[TimeSeries, TimeSeries]]:
    """
    Returns a list of (training_set, validation_set) pairs for backtesting.

    .. todo: this is expanding training window, implement optional sliding window

    :param series: The full time series needs to be split
    :param start: the start time of the earliest validation set
    :param nr_points_val: the number of points in the validation sets
    :param nr_steps_iter: the number of time steps to iterate between the successive validation sets
    :return: a list of (training_set, validation_set) pairs
    """

    assert start in series, 'The provided start timestamp is not in the time series.'
    assert start != series.end_time(), 'The provided start timestamp is the last timestamp of the time series'
    # TODO: maybe also check that valset_duration >= series frequency

    curr_val_start: pd.Timestamp = start

    def _get_train_val_and_increase_pointer() -> Tuple[TimeSeries, TimeSeries]:
        nonlocal curr_val_start

        train_series, val_series_all = series.split_after(curr_val_start)
        val_series = val_series_all.slice_n_points_after(val_series_all.start_time(), nr_points_val)

        curr_val_start = curr_val_start + nr_steps_iter * series.freq()
        return train_series, val_series

    series_pairs = []
    curr_train_series, curr_val_series = _get_train_val_and_increase_pointer()

    while len(curr_val_series) >= nr_points_val:
        series_pairs.append((curr_train_series, curr_val_series))
        curr_train_series, curr_val_series = _get_train_val_and_increase_pointer()

    return series_pairs


def backtest_autoregressive_model(model: 'AutoRegressiveModel',
                                  train_val_series: List[Tuple['TimeSeries', 'TimeSeries']],
                                  eval_fn: Callable[['TimeSeries', 'TimeSeries'], Any]) -> List[Any]:
    """
    Provides a back-testing environment.

    Going through the list 'train_val_series`, this function successively trains a chosen auto-regressive model `model`
    on the training TimeSeries provided, compares the prediction with the validation TimeSeries provided using
    `eval_fn`.

    The results are stored in a list, returned to the user.

    .. todo: option for point-prediction

    :param train_val_series: A list of tuples (train, val) to be passed to the model `model`.
    :param model: The chosen model to use.
    :param eval_fn: The evaluation criteria.
    :return: A list of results obtain via the process described above.
    """

    results = []
    for train, val in train_val_series:
        model.fit(train)
        pred = model.predict(len(val))
        results.append(eval_fn(val, pred))
    return results



from typing import NewType, List, Tuple
from ..timeseries import TimeSeries

# The features used by regressive model
RegrFeatures = NewType('RegrFeatures', List[TimeSeries])

# The "dataset" (features + target) used by regressive model
RegrDataset = NewType('RegrDataset', Tuple[RegrFeatures, TimeSeries])


from u8timeseries.timeseries import TimeSeries
import matplotlib.pyplot as plt
import numpy as np
from typing import Tuple, List
from scipy.interpolate import interp1d


def na_ratio(ts: 'TimeSeries') -> float:
    """
    Computes the ratio of missing values in percent.

    :param ts: The TimeSeries to check for missing values.
    :return: A float, the percentage of missing values in `ts`.
    """

    return 100 * ts.pd_series().isnull().sum() / len(ts)


def fillna_val(ts: 'TimeSeries', fill: float = 0) -> 'TimeSeries':
    """
    Fills the missing values of `ts` with only the value provided (default zeroes).

    :param ts: The TimeSeries to check for missing values.
    :param fill: the value used to replace the missing values.
    :return: A TimeSeries, `ts` with al missing values set to 0.
    """
    return TimeSeries.from_times_and_values(ts.time_index(), ts.pd_series().fillna(value=fill))


def nan_structure_visual(ts: 'TimeSeries', plot: bool = True) -> np.ndarray:
    """
    Plots the indicator function of missing values of `ts`.

    Missing values have value 1 and non-missing values are 0.

    :param ts: The TimeSeries to check.
    :param plot: Boolean to choose to plot the result or not.
    :return: An numpy array containing the indices where the missing values are.
    """

    nans = np.isnan(ts.values())

    if plot: # TODO: find a better way to visualize NaN data
        plt.scatter(np.arange(len(nans)), nans)
        plt.yticks([0, 1], ["Non-missing", "NaN"])
        plt.show()

    return np.where(nans)[0]


def change_of_state(ts: 'TimeSeries') -> Tuple[int, List]:
    """
    Determines the indices where `ts` changes from missing values to non-missing and vice-versa.

    :param ts: The TimeSeries to analyze.
    :return: A tuple (a, l), where a is the number of change of states and l is the list of indices where \
    changes of states occur.
    """

    a = np.where(np.diff(ts.pd_series().isna()))

    return len(a[0]), list(a[0])


def auto_fillna(ts: 'TimeSeries', first: float = None, last: float = None,
                interpolate: str = 'linear', **kwargs) -> 'TimeSeries':
    """
    This function automatically fills the missing value in the TimeSeries `ts`, assuming they are represented by np.nan.

    The rules for completion are given below.

    Missing values at the beginning are filled with constant value `first`. Defaults to backwards-fill.
    Missing values at the end are filled with constant value `last`. Defaults to forward-fill.
    Missing values between to numeric values are set using the interpolation wrapper of pandas with `method`.
    Defaults to linear interpolation.

    Add the option `fill_value` to 'extrapolate' to fill the missing values at the beginning and the end with
    the regression function computed. Must set `first` and `last` to None

    .. todo: be more flexible on the filling methods.

    :param ts: A TimeSeries `ts`.
    :param first: The value to use for filling the beginning of the TimeSeries. Defaults to first known value in `ts`.
    :param last: The value to use for filling the ending of the TimeSeries. Defaults to last known value in `ts`.
    :param interpolate: The function used for filling the middle of the TimeSeries. Defaults to linear interpolation.
    :return: A new TimeSeries with all missing values filled according to the rules above.
    """

    # We compute the number of times entries of the TimeSeries go from missing to numeric and vice-versa
    arr = change_of_state(ts)[1]

    if len(arr) == 0:
        return ts

    ts_temp = ts.pd_series()

    # if first value is missing and `first` is specified, fill values
    if np.isnan(ts.values()[0]) and first is not None:
        ts_temp[:arr[0] + 1] = first

    # if last value is missing and `last` is specified, fill values
    if np.isnan(ts.values()[-1]) and last is not None:
        ts_temp[arr[-1]+1:] = last

    # pandas interpolate wrapper, with chosen `method`
    ts_temp.interpolate(method=interpolate, inplace=True, limit_direction='both', limit=len(ts_temp), **kwargs)

    return TimeSeries.from_times_and_values(ts.time_index(), ts_temp.values)



from .regressive_model import RegressiveModel
from ..timeseries import TimeSeries
from typing import List
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression


class StandardRegressiveModel(RegressiveModel):

    def __init__(self, train_n_points, model=LinearRegression(n_jobs=-1, fit_intercept=False)):
        """
        Simple regression based on other fit() predict() models (e.g., from sklearn).

        :param train_n_points: The number of most recent points from the training time series that
                               will be used to train the regressive model. If the provided training series
                               contain fewer points, they will all be used for training.
        :param model: The actual regressive model. It must contain fit() and predict() methods.
        """
        super(StandardRegressiveModel, self).__init__()
        assert callable(getattr(model, "fit", None)), 'Provided model object must have a fit() method'
        assert callable(getattr(model, "predict", None)), 'Provided model object must have a predict() method'

        self.train_n_points = train_n_points
        self.model = model

    @staticmethod
    def _get_features_matrix_from_series(features: List[TimeSeries]):
        return np.array([s.values() for s in features]).T  # (n_samples x n_features)

    def fit(self, train_features: List[TimeSeries], train_target: TimeSeries):
        # Get (at most) the last [train_n_points] of each series
        last_train_ts = train_features[0].end_time()
        last_n_points_features = [s.slice_n_points_before(last_train_ts, self.train_n_points) for s in train_features]
        last_n_points_target = train_target.slice_n_points_before(last_train_ts, self.train_n_points)

        super().fit(last_n_points_features, last_n_points_target)

        self.model.fit(self._get_features_matrix_from_series(last_n_points_features), last_n_points_target.values())

    def predict(self, features: List[TimeSeries]):
        super().predict(features)
        y = self.model.predict(self._get_features_matrix_from_series(features))
        return TimeSeries(pd.Series(y, index=features[0].time_index()))



import unittest
import pandas as pd
import numpy as np

from ..timeseries import TimeSeries
from u8timeseries import Prophet, KthValueAgoBaseline, ExponentialSmoothing, TimeSeries, Arima, AutoArima
from u8timeseries.models.theta import Theta

class ModelsTestCase(unittest.TestCase):

    # forecasting horizon used in predictions
    forecasting_horizon = 5

    # dummy timeseries for autoregression forecasting
    times = pd.date_range('20000101', '20000130')
    values = np.sin(range(len(times))) + np.array([2.0] * len(times))
    ts: TimeSeries = TimeSeries.from_times_and_values(times, values)


    def test_autoregressive_models_runnability(self):
        models = [
            ExponentialSmoothing(), 
            Prophet(),
            Arima(1, 1, 1),
            AutoArima(),
            KthValueAgoBaseline(),
            Theta()
        ]

        for model in models:
            model.fit(self.ts)
            prediction = model.predict(self.forecasting_horizon)
            self.assertTrue(len(prediction) == self.forecasting_horizon)



from ..timeseries import TimeSeries
import numpy as np
from statsmodels.tsa.stattools import acf
from scipy.stats import norm
from statsmodels.tsa.seasonal import seasonal_decompose
import matplotlib.pyplot as plt
from typing import Tuple
import math


def check_seasonality(ts: 'TimeSeries', m: int = None, max_lag: int = 24, alpha: float = 0.05):
    """
    Returns whether the TimeSeries `ts` is seasonal with period `m` or not.

    If [m] is None, we work under the assumption that there is a unique seasonality period, which is inferred
    from the Auto-correlation Function (ACF).

    .. todo: Generalize to multiple seasonality periods.

    :param ts: The TimeSeries to check for seasonality.
    :param m: The seasonality period to check.
    :param max_lag: The maximal lag allowed in the ACF.
    :param alpha: The desired confidence level (default 5%).
    :return: A tuple (season, m), where season is a boolean indicating whether the TimeSeries has seasonality or not
             and m is the seasonality period.
    """
    if m is not None and (m < 2 or not isinstance(m, int)):
        raise ValueError('m must be an integer greater than 1.')

    if m is not None and m > max_lag:
        raise ValueError('max_lag must be greater than or equal to m.')

    n_unique = np.unique(ts.values()).shape[0]

    if n_unique == 1:  # Check for non-constant TimeSeries
        return False, 0

    r = acf(ts.values(), nlags=max_lag)  # In case user wants to check for seasonality higher than 24 steps.

    gradient = np.gradient(r)
    gradient_signs_changes = np.diff(np.sign(gradient))

    # Tries to infer seasonality from Auto-Correlation Function if no value of m has been provided.
    # We look for the first positive significant local maximum of the ACF by checking the sign changes
    # in the gradient.

    # Local maximum is indicated by signs_change == -2.
    if len(np.nonzero((gradient_signs_changes == -2))[0]) == 0:
        print('The ACF has no local maximum for m < max_lag = {}.'.format(max_lag))
        return False, 0

    # Building a list of candidates for local maximum.
    candidates = np.nonzero((gradient_signs_changes == -2))[0].tolist()

    # If a -2 value appears in gradient_signs_changes at index i, then the local
    # maximum of r occurs either at index i or i+1. We check manually and change the candidates accordingly.
    candidates = [i if r[i] >= r[i + 1] else i + 1 for i in candidates]

    if m is not None:
        # Check for local maximum when m is user defined.
        test = m not in candidates

        if test:
            return False, m

        candidates = [m]

    # Remove r[0], the auto-correlation at lag order 0, that introduces bias.
    r = r[1:]

    # The non-adjusted upper limit of the significance interval.
    band_upper = r.mean() + norm.ppf(1 - alpha / 2) * r.var()

    # Significance test, stops at first admissible value. The two '-1' below
    # compensate for the index change due to the restriction of the original r to r[1:].
    for candidate in candidates:
        stat = _bartlett_formula(r, candidate - 1, len(ts))
        if r[candidate - 1] > stat * band_upper:
            return True, candidate
    return False, 0


def _bartlett_formula(r, m, length) -> float:
    """
    Computes the standard error of `r` at order `m` with respect to `length` according to Bartlett's formula.

    :param r: The array whose standard error is to be computed.
    :param m: The order of the standard error.
    :param length: The size of the underlying sample to be used.
    :return: The standard error of `r` with order `m`.
    """
    if m == 1:
        return math.sqrt(1/length)
    else:
        return math.sqrt((1 + 2 * sum(map(lambda x: x ** 2, r[:m-1]))) / length)


def extract_trend_and_seasonality(ts: 'TimeSeries', freq: int = None, model: str = 'multiplicative') \
                                                                -> Tuple['TimeSeries', 'TimeSeries']:
    """
    Extracts trend and seasonality from the TimeSeries `ts` using statsmodels.seasonal_decompose.

    :param ts: The TimeSeries to study.
    :param freq: The seasonality period to use.
    :param model: The type of decomposition to use ('additive' or 'multiplicative').
    :return: A tuple (trend, seasonal) of TimeSeries containing the trend and seasonality respectively.
    """

    decomp = seasonal_decompose(ts.pd_series(), freq=freq, model=model, extrapolate_trend='freq')

    season = TimeSeries.from_times_and_values(ts.time_index(), decomp.seasonal)
    trend = TimeSeries.from_times_and_values(ts.time_index(), decomp.trend)

    return trend, season


def remove_from_series(ts: 'TimeSeries', other: 'TimeSeries', model: str) -> 'TimeSeries':
    """
    Removes the TimeSeries `other` from the TimeSeries `ts` according to the specified model.

    :param ts: The TimeSeries to be modified.
    :param other: The TimeSeries to remove.
    :param model: The type of model considered (either 'additive' or 'multiplicative').
    :return: A TimeSeries `new_ts`, defined by removing `other` from `ts`.
    """
    if (other.values() < 1e-6).any() and model == 'multiplicative':
        print("WARNING indexes equal to zero, cannot remove for a multiplicative model.")
        return ts
    else:
        if model == 'multiplicative':
            new_ts = ts / other
        else:
            new_ts = ts - other

    return new_ts


def remove_seasonality(ts: 'TimeSeries', freq: int = None, model: str = 'multiplicative') \
                                                        -> 'TimeSeries':
    """
    Adjusts the TimeSeries `ts` for a seasonality of order `frequency` using the `model` decomposition.

    :param ts: The TimeSeries to adjust.
    :param freq: The seasonality period to use.
    :param model: The type of decomposition to use (additive or multiplicative).
    :return: A tuple (new_ts, seasonality) of TimeSeries where new_ts is the adjusted original TimeSeries and
            seasonality is a TimeSeries containing the removed seasonality.
    """

    _, seasonality = extract_trend_and_seasonality(ts, freq, model)

    new_ts = remove_from_series(ts, seasonality, model)

    return new_ts


def remove_trend(ts: 'TimeSeries', model='multiplicative') -> 'TimeSeries':
    """
    Adjusts the TimeSeries `ts` for a trend using the `model` decomposition.

    :param ts: The TimeSeries to adjust.
    :param model: The type of decomposition to use (additive or multiplicative).
    :return: A tuple (new_ts, trend) of TimeSeries where new_ts is the adjusted original TimeSeries and
            seasonality is a TimeSeries containing the removed trend.
    """
    trend, _ = extract_trend_and_seasonality(ts, model=model)

    new_ts = remove_from_series(ts, trend, model)

    return new_ts


def plot_acf(ts: 'TimeSeries', m: int = None, max_lag: int = 24, alpha: float = 0.05,
             fig_size: Tuple[int, int] = (10, 5)):
    """
    Plots the ACF of `ts`, highlighting it at lag `m`, with corresponding significance interval.

    :param ts: The TimeSeries whose ACF should be plotted.
    :param m:  The lag to highlight.
    :param max_lag: The maximal lag order to consider.
    :param alpha: The confidence interval.
    :param fig_size: The size of the figure to be displayed.
    """
    r = acf(ts.values(), nlags=max_lag)  # , alpha=alpha) and confint as output too

    # Computes the confidence interval at level alpha for all lags.
    stats = []
    for i in range(1, max_lag+1):
        stats.append(_bartlett_formula(r[1:], i, len(ts)))

    plt.figure(figsize=fig_size)

    for i in range(len(r)):
        plt.plot((i, i), (0, r[i]), color=('red' if m is not None and i == m else 'black'), lw=.5)

    upp_band = r[1:].mean() + norm.ppf(1 - alpha/2) * r[1:].var()
    acf_band = [upp_band * stat for stat in stats]

    plt.fill_between(np.arange(1, max_lag+1), acf_band, [-x for x in acf_band], color='blue', alpha=.25)
    plt.plot((0, max_lag+1), (0, 0), color='black')

    plt.show()



from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
from ..timeseries import TimeSeries
from typing import Optional


class AutoRegressiveModel(ABC):
    """
    Base class for implementation of Auto-regressive models.

    This is a base class for various implementation of uni-variate time series forecasting models.
    These models predict future values of one time series using no other data.
    """

    @abstractmethod
    def __init__(self):
        # Stores training date information:
        self.training_series: TimeSeries = None

        # state
        self._fit_called = False

    @abstractmethod
    def fit(self, series: TimeSeries) -> None:
        self.training_series = series
        self._fit_called = True

    @abstractmethod
    def predict(self, n: int) -> TimeSeries:
        """
        :return: A TimeSeries containing the `n` next points, starting after the end of the training time series.
        """
        assert self._fit_called, 'fit() must be called before predict()'

    def _generate_new_dates(self, n: int):
        """
        Generate `n` new dates after the end of the training set
        """
        new_dates = [self.training_series.time_index()[-1] + (i * self.training_series.freq()) for i in range(1, n+1)]
        return pd.DatetimeIndex(new_dates, freq=self.training_series.freq_str())

    def _build_forecast_series(self, points_preds: np.ndarray,
                               lower_bound: Optional[np.ndarray] = None,
                               upper_bound: Optional[np.ndarray] = None):

        time_index = self._generate_new_dates(len(points_preds))

        return TimeSeries.from_times_and_values(time_index, points_preds, lower_bound, upper_bound)



import unittest
import pandas as pd
import numpy as np

from u8timeseries import TimeSeries
from u8timeseries.models.statistics import check_seasonality


class TimeSeriesTestCase(unittest.TestCase):

    def test_check_seasonality(self):
        pd_series = pd.Series(range(50), index=pd.date_range('20130101', '20130219'))
        pd_series = pd_series.map(lambda x: np.sin(x*np.pi / 3 + np.pi/2))
        series = TimeSeries(pd_series)

        self.assertEqual((True, 6), check_seasonality(series))
        self.assertEqual((False, 3), check_seasonality(series, m=3))




import statsmodels.tsa.holtwinters as hw
from ..timeseries import TimeSeries
from .autoregressive_model import AutoRegressiveModel
import numpy as np
import math
from .statistics import check_seasonality, extract_trend_and_seasonality, remove_seasonality


class Theta(AutoRegressiveModel):
    """
    An implementation of the Theta method with variable value of the `theta` parameter.

    :param theta: User-defined value for the theta parameter. Default to 0.

    .. todo: Implement OTM: Optimized Theta Method (https://arxiv.org/pdf/1503.03529.pdf)
    .. todo: From the OTM, set theta_2 = 2-theta_1 to recover our generalization - but we have an explicit formula.
    .. todo: Try with something different than SES? They do that in the paper.
    """
    def __init__(self, theta: int = 0):
        super().__init__()

        self.model = None
        self.coef = 1
        self.alpha = 1
        self.length = 0
        self.theta = theta
        self.is_seasonal = False
        self.seasonality = None
        self.season_period = 0

        # Remark on the values of the theta parameter:
        # - if theta = 1, then the theta method restricts to a simple exponential smoothing (SES)
        # - if theta = 2, the formula for self.coef below fails, hence it is forbidden.

        if self.theta == 2:
            raise ValueError('The parameter theta cannot be equal to 2.')

    def fit(self, ts, season_period: int = None):
        """
        Fits the Theta method to the TimeSeries `ts`.


        The model decomposition is defined by the parameters `theta`, and the TimeSeries `ts`
        is de-seasonalized according to `season_period`.

        :param ts: The TimeSeries to fit.
        :param season_period: User-defined seasonality period. Default to None.
        """
        super().fit(ts)

        self.length = len(ts)

        # Check for statistical significance of user-defined season period
        # or infers season_period from the TimeSeries itself.
        self.is_seasonal, self.season_period = check_seasonality(ts, season_period)

        new_ts = ts

        # Store and remove seasonality effect if there is any.
        if self.is_seasonal:
            _, self.seasonality = extract_trend_and_seasonality(ts, self.season_period)
            new_ts = remove_seasonality(ts, self.season_period)

        # SES part of the decomposition.
        self.model = hw.SimpleExpSmoothing(new_ts.values()).fit()

        # Linear Regression part of the decomposition. We select the degree one coefficient.
        b_theta = np.polyfit(np.array([i for i in range(0, self.length)]), (1.0 - self.theta) * new_ts.values(), 1)[0]

        # Normalization of the coefficient b_theta.
        self.coef = b_theta / (2.0 - self.theta)

        self.alpha = self.model.params["smoothing_level"]

    def predict(self, n: int) -> 'TimeSeries':
        """
        Uses the fitted model to predict `n` values in the future.

        :param n: The length of the horizon to predict over.
        :return: A new TimeSeries containing the `n` prediction points.
        """
        super().predict(n)

        # Forecast of the SES part.
        forecast = self.model.forecast(n)

        # Forecast of the Linear Regression part.
        drift = self.coef * np.array([i + (1 - (1 - self.alpha) ** self.length) / self.alpha for i in range(0, n)])

        # Combining the two forecasts
        forecast += drift

        # Re-apply the seasonal trend of the TimeSeries
        if self.is_seasonal:

            replicated_seasonality = np.tile(self.seasonality.pd_series()[-self.season_period:],
                                             math.ceil(n / self.season_period))[:n]
            forecast *= replicated_seasonality

        return self._build_forecast_series(forecast)



from .autoregressive_model import AutoRegressiveModel
from ..timeseries import TimeSeries
import numpy as np


class KthValueAgoBaseline(AutoRegressiveModel):
    """
    A baseline model that always predict value of `k` time steps ago.


    More precisely, at last know time value t, the prediction for t + i is given by the value at time t - k + i.

    :param k: An integer, determines how far to fetch the prediction value.
    """

    def __init__(self, k: int = 1):
        super().__init__()
        self.last_k_vals = None
        self.K = k

    def __str__(self):
        return '{} value ago baseline'.format(self.K)

    def fit(self, series: 'TimeSeries'):
        super().fit(series)
        assert len(series) >= self.K, 'The time series has to contain at least K={} points'.format(self.K)
        self.last_k_vals = series.values()[-self.K:]

    def predict(self, n: int):
        super().predict(n)
        forecast = np.array([self.last_k_vals[i % self.K] for i in range(n)])
        return self._build_forecast_series(forecast)



from u8timeseries.models.autoregressive_model import AutoRegressiveModel
import pandas as pd

import fbprophet


class Prophet(AutoRegressiveModel):
    """
    Implementation of the Prophet model.

    Currently just a wrapper around the fbprophet implementation.

    :param country_holidays: An optional country code, for which holidays can be taken into account by Prophet.

                             See: https://github.com/dr-prodigy/python-holidays

                             In addition to those countries, Prophet includes holidays for these
                             countries: Brazil (BR), Indonesia (ID), India (IN), Malaysia (MY), Vietnam (VN),
                             Thailand (TH), Philippines (PH), Turkey (TU), Pakistan (PK), Bangladesh (BD),
                             Egypt (EG), China (CN), and Russian (RU).
    :param weekly_seasonality:
    :param daily_seasonality:
    """

    def __init__(self, weekly_seasonality=False, daily_seasonality=False, country_holidays: str = None):

        super().__init__()

        self.country_holidays = country_holidays
        self.weekly_seasonality = weekly_seasonality
        self.daily_seasonality = daily_seasonality
        self.model = None

    def __str__(self):
        return 'Prophet'

    def fit(self, series):
        super().fit(series)

        in_df = pd.DataFrame(data={
            'ds': series.time_index(),
            'y': series.values()
        })

        # TODO: user-provided seasonalities, or "auto" based on stepduration
        self.model = fbprophet.Prophet(weekly_seasonality=self.weekly_seasonality,
                                       daily_seasonality=self.daily_seasonality)

        # Input built-in country holidays
        if self.country_holidays is not None:
            self.model.add_country_holidays(self.country_holidays)

        self.model.fit(in_df)

    def predict(self, n):
        super().predict(n)
        new_dates = self._generate_new_dates(n)
        new_dates_df = pd.DataFrame(data={'ds': new_dates})

        predictions = self.model.predict(new_dates_df)

        forecast = predictions['yhat'][-n:].values
        conf_lo = predictions['yhat_lower'][-n:].values
        conf_hi = predictions['yhat_upper'][-n:].values
        return self._build_forecast_series(forecast, conf_lo, conf_hi)



import unittest
import pandas as pd
import numpy as np

from ..timeseries import TimeSeries
from u8timeseries import StandardRegressiveModel

class RegressionModelsTestCase(unittest.TestCase):

    # number of data points used for training
    regression_window = 5

    # dummy feature and target timeseries for regression
    times = pd.date_range('20000101', '20000130')
    target_values = np.sin(range(len(times))) + np.array([2.0] * len(times))
    feature_values = np.array(range(len(times)), dtype=float)
    target_ts: TimeSeries = TimeSeries.from_times_and_values(times, target_values)
    feature_ts: TimeSeries = TimeSeries.from_times_and_values(times, feature_values)

    
    def test_regressive_models_runnability(self):
        models = [
            StandardRegressiveModel(self.regression_window)
        ]

        for model in models:
            # training and predicting on same features, since only runnability is tested
            model.fit([self.feature_ts], self.target_ts)
            prediction = model.predict([self.feature_ts])
            self.assertTrue(len(prediction) == len(self.feature_ts))



import unittest
import pandas as pd
import numpy as np

from ..timeseries import TimeSeries
from u8timeseries.utils.missing_values import auto_fillna


class MissingValuesTestCase(unittest.TestCase):

    time = pd.date_range('20130101', '20130130')
    lin = [float(i) for i in range(len(time))]
    cub = [float(i-4)**2 for i in range(len(time))]
    series1: TimeSeries = TimeSeries.from_times_and_values(time, np.array([2.0]*len(time)))
    series2: TimeSeries = TimeSeries.from_times_and_values(time, np.array(lin))
    series3: TimeSeries = TimeSeries.from_times_and_values(time, np.array([10]*10 + lin[-20:]))
    series4: TimeSeries = TimeSeries.from_times_and_values(time, np.array(lin[:20] + [19]*10))
    series5: TimeSeries = TimeSeries.from_times_and_values(time, np.array(cub))
    series6: TimeSeries = TimeSeries.from_times_and_values(time, [0]*2 + cub[2:-2] + [-1]*2)

    def test_fill_constant(self):
        seriesA: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                    np.array([np.nan] * 5 + [2.0] * 5 + [np.nan] * 5 + [2.0] * 10 + [np.nan] * 5))

        # Check that no changes are made if there are no missing values
        self.assertEqual(self.series1, auto_fillna(self.series1))

        # Check that a constant function is filled to a constant function
        self.assertEqual(self.series1, auto_fillna(seriesA))

    def test_linear(self):
        seriesB: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                                               np.array(self.lin[:10] + [np.nan]*10 + self.lin[-10:]))

        # Check for linear interpolation part
        self.assertEqual(self.series2, auto_fillna(seriesB))

    def test_bfill(self):
        seriesC: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                                               np.array([np.nan] * 10 + self.lin[-20:]))

        # Check that auto-backfill works properly
        self.assertEqual(self.series3, auto_fillna(seriesC))
        self.assertNotEqual(self.series3, auto_fillna(seriesC, first=2))

    def test_ffil(self):
        seriesD: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                                               np.array(self.lin[:20] + [np.nan] * 10))

        self.assertEqual(self.series4, auto_fillna(seriesD))
        self.assertNotEqual(self.series4, auto_fillna(seriesD, last=20))

    def test_fill_quad(self):
        seriesE: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                                               np.array(self.cub[:10] + [np.nan]*10 + self.cub[-10:]))
        seriesF: TimeSeries = TimeSeries.from_times_and_values(self.time,
                                                               np.array([np.nan]*2 + self.cub[2:10] +
                                                                        [np.nan]*10 + self.cub[-10:-2] + [np.nan]*2))

        self.assertEqual(self.series5, round(auto_fillna(seriesE, interpolate='quadratic'), 7))
        self.assertEqual(self.series6, round(auto_fillna(seriesF, first=0, last=-1, interpolate='quadratic'), 2))
        # extrapolate values outside
        self.assertEqual(self.series5,
                         round(auto_fillna(seriesF, interpolate='quadratic', fill_value='extrapolate'), 2))




import unittest
import pandas as pd
import numpy as np

from ..timeseries import TimeSeries
from u8timeseries.utils.timeseries_generation import (
    constant_timeseries, 
    linear_timeseries, 
    periodic_timeseries,
    white_noise_timeseries, 
    random_walk_timeseries, 
    us_holiday_timeseries
)

class TimeSeriesGenerationTestCase(unittest.TestCase):

    def test_constant_timeseries(self):

        # testing parameters
        length = 100
        value = 5

        # testing for constant value
        constant_ts = constant_timeseries(value=value, length=length)
        value_set = set(constant_ts._series.values)
        self.assertTrue(len(value_set) == 1)

    def test_linear_timeseries(self):

        # testing parameters
        length = 100
        start_value = 5
        value_delta = 3

        # testing for correct start and end values
        linear_ts = linear_timeseries(start_value=start_value, value_delta=value_delta, length=length)
        self.assertEqual(linear_ts.values()[0], start_value)
        self.assertEqual(linear_ts.values()[-1], start_value + (length - 1) * value_delta)

    def test_periodic_timeseries(self):

        # testing parameters
        length = 100
        amplitude = 5
        y_offset = -3

        # testing for correct value range
        periodic_ts = periodic_timeseries(length=length, amplitude=amplitude, y_offset=y_offset)
        self.assertTrue((periodic_ts <= y_offset + amplitude).all())
        self.assertTrue((periodic_ts >= y_offset - amplitude).all())

    def test_white_noise_timeseries(self):

        # testing parameters
        length = 100

        # testing for correct length
        white_noise_ts = white_noise_timeseries(length=length)
        self.assertEqual(len(white_noise_ts), length)

    def test_random_walk_timeseries(self):

        # testing parameters
        length = 100

        # testing for correct length
        random_walk_ts = random_walk_timeseries(length=length)
        self.assertEqual(len(random_walk_ts), length)

    def test_us_holiday_timeseries(self):

        # testing parameters
        length = 30
        start_date = pd.Timestamp('20201201')

        # testing for christmas and non-holiday
        us_holiday_ts = us_holiday_timeseries(length=length, start_date=start_date)
        self.assertEqual(us_holiday_ts._series.at[pd.Timestamp('20201225')], 1)
        self.assertEqual(us_holiday_ts._series.at[pd.Timestamp('20201210')], 0)



from abc import ABC, abstractmethod
from ..timeseries import TimeSeries
from typing import List


class RegressiveModel(ABC):
    """
    This is a base class for various implementations of multi-variate models - models predicting time series
    from one or several time series. It also allows to do ensembling.

    TODO: Extend this to a "DynamicRegressiveModel" class, which acts on List[List[TimeSeries]].
    TODO: The first List[] would contain time-sliding lists of time series, letting the model
    TODO: be able to learn how to change weights over time. When len() of outer List[] is 0 it's a particular case
    """

    @abstractmethod
    def __init__(self):
        # Stores training date information:
        self.train_features: List[TimeSeries] = None
        self.train_target: TimeSeries = None

        # state
        self._fit_called = False

    @abstractmethod
    def fit(self, train_features: List[TimeSeries], train_target: TimeSeries) -> None:
        assert len(train_features) > 0, 'Need at least one feature series'
        assert all([s.has_same_time_as(train_target) for s in train_features]), 'All provided time series must ' \
                                                                                'have the same time index'
        self.train_features = train_features
        self.train_target = train_target
        self._fit_called = True

    @abstractmethod
    def predict(self, features: List[TimeSeries]) -> TimeSeries:
        """
        :return: A TimeSeries containing the prediction obtained from [features], of same length as [features]
        """
        assert self._fit_called, 'fit() must be called before predict()'
        assert len(features) == len(self.train_features), 'Provided features must have same dimensionality as ' \
                                                          'training features. There were {} training features and ' \
                                                          'the function has been called with {} features' \
                                                          .format(len(self.train_features), len(features))

    def residuals(self) -> TimeSeries:
        """
        :return: a time series of residuals (absolute errors of the model on the training set)
        """
        assert self._fit_called, 'fit() must be called before residuals()'

        train_pred = self.predict(self.train_features)
        return abs(train_pred - self.train_target)



from .regressive_model import RegressiveModel
from ..timeseries import TimeSeries
from typing import List


class SequentialAggregation(RegressiveModel):
    """
    TODO
    https://s3.amazonaws.com/academia.edu.documents/45700981/Forecasting_electricity_consumption_by_a20160517-4457-1a8m354.pdf?AWSAccessKeyId=AKIAIWOWYYGZ2Y53UL3A&Expires=1552341840&Signature=x7vegA6suvqbvDuyLiaK%2BKCt23s%3D&response-content-disposition=inline%3B%20filename%3DForecasting_electricity_consumption_by_a.pdf

    https://link.springer.com/article/10.1007/s10994-012-5314-7

    https://hal.inria.fr/tel-01697133v2/document (Section 1.2)
    """
    def __init__(self):
        super().__init__()

    def fit(self, train_features: List[TimeSeries], train_target: TimeSeries):
        super().fit(train_features, train_target)

    def predict(self, features: List[TimeSeries]):
        super().predict(features)


import numpy as np
import pandas as pd
import math
from pandas.tseries.offsets import DateOffset
from pandas.tseries.frequencies import to_offset
from pandas.tseries.holiday import USFederalHolidayCalendar

from u8timeseries.timeseries import TimeSeries



def constant_timeseries(value: float = 0, length: int = 10, offset: str = 'D',
                        start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a constant TimeSeries with the given value, length, start date and offset.

    :param value: The constant value that the TimeSeries object will assume at every index.
    :param length: The length of the returned TimeSeries.
    :param offset: The time differene between two adjacent entries in the returned TimeSeries. A DateOffset alias is expected;
                   see: https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: A constant TimeSeries with value 'value'.
    """

    times = pd.date_range(periods=length, freq=offset, start=start_date)
    values = np.full(length, value)

    return TimeSeries.from_times_and_values(times, values)


def linear_timeseries(start_value: float = 0, value_delta: float = 1, length: int = 10, offset: str = 'D',
                      start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a TimeSeries with a starting value of 'start_value' that increases by 'value_delta' at each step.
    The last entry of the time series will be equal to 'start_value' + ('length' - 1) * 'value_delta'.

    :param start_value: The value of the first entry in the TimeSeries.
    :param value_delta: The difference in value between to adjacent entries in the TimeSeries.
    :param length: The length of the returned TimeSeries.
    :param offset: The time differene between two adjacent entries in the returned TimeSeries. A DateOffset alias is expected.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: A linear TimeSeries with gradient 'value_delta'.
    """

    times = pd.date_range(periods=length, freq=offset, start=start_date)
    values = np.linspace(start_value, start_value + (length - 1) * value_delta, length)

    return TimeSeries.from_times_and_values(times, values)


def periodic_timeseries(frequency: float = 0.1, amplitude: float = 1, phase: float = 0, y_offset: float = 0,
                        length: int = 10, offset: str = 'D',
                        start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a TimeSeries with a sinusoidal value progression with a given frequency, amplitude, phase and y offset.

    :param frequency: The number of periods that take place within one time unit given in 'offset'.
    :param amplitude: The maximum  difference between any value of the returned TimeSeries and 'y_offset'.
    :param phase: The relative position within one period of the first value of the returned TimeSeries (in radians).
    :param y_offset: The shift of the sine function along the y axis.
    :param length: The length of the returned TimeSeries.
    :param offset: The time differene between two adjacent entries in the returned TimeSeries. A DateOffset alias is expected.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: A sinusoidal TimeSeries parametrized as indicated above.
    """

    times = pd.date_range(periods=length, freq=offset, start=start_date)
    values = np.array(range(length), dtype=float)
    f = np.vectorize(lambda x: amplitude * math.sin(2 * math.pi * frequency * x + phase) + y_offset)
    values = f(values)

    return TimeSeries.from_times_and_values(times, values)


def white_noise_timeseries(length: int = 10, offset: str = 'D', mean: float = 0, std: float = 1,
                           start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a white noise TimeSeries by sampling a gaussian distribution with mean 'mean' and 
    standard deviation 'std'. Each value represents an indipendent sample of the distribution.

    :param mean: The mean of the gaussian distribution that is sampled at each step.
    :param std: The standard deviation of the gaussian distribution that is sampled at each step.
    :param length: The length of the returned TimeSeries.
    :param offset: The time differene between two adjacent entries in the returned TimeSeries. A DateOffset alias is expected.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: A white noise TimeSeries created as indicated above.
    """

    times = pd.date_range(periods=length, freq=offset, start=start_date)
    values = np.random.normal(mean, std, size=length)

    return TimeSeries.from_times_and_values(times, values)


def random_walk_timeseries(length: int = 10, offset: str = 'D', mean: float = 0, std: float = 1,
                           start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a random walk TimeSeries by sampling a gaussian distribution with mean 'mean' and 
    standard deviation 'std'. The first value is one such random sample. Every subsequent value
    is equal to the previous value plus a random sample.

    :param mean: The mean of the gaussian distribution that is sampled at each step.
    :param std: The standard deviation of the gaussian distribution that is sampled at each step.
    :param length: The length of the returned TimeSeries.
    :param offset: The time differene between two adjacent entries in the returned TimeSeries. A DateOffset alias is expected.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: A random walk TimeSeries created as indicated above.
    """

    times = pd.date_range(periods=length, freq=offset, start=start_date)
    values = [np.random.normal(mean, std)]
    while (len(values) < length):
        values.append(values[-1] + np.random.normal(mean, std))

    return TimeSeries.from_times_and_values(times, values)


def us_holiday_timeseries(length: int = 10, start_date: pd.Timestamp = pd.Timestamp('2000-01-01')) -> 'TimeSeries':
    """
    Creates a binary TimeSeries that equals 1 at every index that corresponds to a US holiday, 
    and 0 otherwise. The frequency of the TimeSeries is daily.

    :param length: The length of the returned TimeSeries.
    :param start_date: The time index of the first entry in the returned TimeSeries.
    :return: Binary TimeSeries for US holidays.
    """

    times = pd.date_range(periods=length, freq='D', start=start_date)
    us_holidays = USFederalHolidayCalendar().holidays()
    values = times.isin(us_holidays).astype(int)
    
    return TimeSeries.from_times_and_values(times, values)




    



import unittest
import numpy as np
import pandas as pd

from ..timeseries import TimeSeries
from u8timeseries.metrics import metrics


class MetricsTestCase(unittest.TestCase):

    pd_series1 = pd.Series(range(10), index=pd.date_range('20130101', '20130110'))
    pd_series2 = pd.Series(np.random.rand(10)*10, index=pd.date_range('20130101', '20130110'))
    pd_series3 = pd.Series(np.sin(np.pi*np.arange(20)/4), index=pd.date_range('20130101', '20130120'))
    series1: TimeSeries = TimeSeries(pd_series1)
    pd_series1[:] = pd_series1.mean()
    series0: TimeSeries = TimeSeries(pd_series1)
    series2: TimeSeries = TimeSeries(pd_series2)
    series3: TimeSeries = TimeSeries(pd_series3)

    def test_zero(self):
        self.assertTrue(np.isnan(metrics.mape(self.series1, self.series1)))

        self.assertTrue(np.isnan(metrics.overall_percentage_error(self.series1-self.series1.mean(),
                                                                  self.series1-self.series1.mean())))

    def test_same(self):
        self.assertEqual(metrics.mape(self.series1+1, self.series1+1), 0)
        self.assertEqual(metrics.mase(self.series1+1, self.series1+1, 1), 0)
        self.assertEqual(metrics.marre(self.series1+1, self.series1+1), 0)
        self.assertEqual(metrics.r2_score(self.series1+1, self.series1+1), 1)
        self.assertEqual(metrics.overall_percentage_error(self.series1+1, self.series1+1), 0)

    def test_r2(self):
        from sklearn.metrics import r2_score
        self.assertEqual(metrics.r2_score(self.series1, self.series0), 0)
        self.assertEqual(metrics.r2_score(self.series1, self.series2),
                         r2_score(self.series1.values(), self.series2.values()))

    def test_marre(self):
        self.assertAlmostEqual(metrics.marre(self.series1, self.series2),
                               metrics.marre(self.series1+100, self.series2+100))

    def test_season(self):
        with self.assertRaises(AssertionError):
            metrics.mase(self.series3, self.series3 * 1.3, 8)





from u8timeseries.models.arima import Arima, AutoArima
from u8timeseries.models.baselines import KthValueAgoBaseline
from u8timeseries.models.exponential_smoothing import ExponentialSmoothing
from u8timeseries.models.prophet import Prophet
from u8timeseries.models.standard_regressive_model import StandardRegressiveModel
from u8timeseries.models.autoregressive_model import AutoRegressiveModel
from u8timeseries.models.theta import Theta
from .timeseries import TimeSeries



import unittest
import pandas as pd
import numpy as np

from ..timeseries import TimeSeries


class TimeSeriesTestCase(unittest.TestCase):

    times = pd.date_range('20130101', '20130110')
    pd_series1 = pd.Series(range(10), index=times)
    pd_series2 = pd.Series(range(5, 15), index=times)
    pd_series3 = pd.Series(range(15, 25), index=times)
    series1: TimeSeries = TimeSeries(pd_series1)
    series2: TimeSeries = TimeSeries(pd_series1, pd_series2, pd_series3)
    series3: TimeSeries = TimeSeries(pd_series2)

    def test_creation(self):
        with self.assertRaises(AssertionError):
            # Index is dateTimeIndex
            TimeSeries(pd.Series(range(10), range(10)))

        with self.assertRaises(AssertionError):
            # Conf interval must be same length as main series
            pd_lo = pd.Series(range(5, 14), index=pd.date_range('20130101', '20130109'))
            TimeSeries(self.pd_series1, pd_lo)

        with self.assertRaises(AssertionError):
            # Conf interval must have same time index as main series
            pd_lo = pd.Series(range(5, 15), index=pd.date_range('20130102', '20130111'))
            TimeSeries(self.pd_series1, pd_lo)

        with self.assertRaises(AssertionError):
            # Conf interval must be same length as main series
            pd_hi = pd.Series(range(5, 14), index=pd.date_range('20130101', '20130109'))
            TimeSeries(self.pd_series1, None, pd_hi)

        with self.assertRaises(AssertionError):
            # Conf interval must have same time index as main series
            pd_lo = pd.Series(range(5, 15), index=pd.date_range('20130102', '20130111'))
            TimeSeries(self.pd_series1, None, pd_lo)

        with self.assertRaises(AssertionError):
            # Main series cannot have date holes
            range_ = pd.date_range('20130101', '20130104').append(pd.date_range('20130106', '20130110'))
            TimeSeries(pd.Series(range(9), index=range_))

        series_test = TimeSeries(self.pd_series1, self.pd_series2, self.pd_series3)

        self.assertTrue(series_test.pd_series().equals(self.pd_series1))
        self.assertTrue(series_test.conf_lo_pd_series().equals(self.pd_series2))
        self.assertTrue(series_test.conf_hi_pd_series().equals(self.pd_series3))

    def test_alt_creation(self):
        with self.assertRaises(AssertionError):
            # Series cannot be lower than three
            index = pd.date_range('20130101', '20130102')
            TimeSeries.from_times_and_values(index, self.pd_series1.values[:2])
        with self.assertRaises(ValueError):
            # all array must have same length
            TimeSeries.from_times_and_values(self.pd_series1.index,
                                             self.pd_series1.values[:-1],
                                             self.pd_series2[:-2],
                                             self.pd_series3[:-1])

        # test if reordering is correct
        rand_perm = np.random.permutation(range(1, 11))
        index = pd.to_datetime(['201301{:02d}'.format(i) for i in rand_perm])
        series_test = TimeSeries.from_times_and_values(index, self.pd_series1.values[rand_perm-1],
                                                       self.pd_series2[rand_perm-1],
                                                       self.pd_series3[rand_perm-1].tolist())

        self.assertTrue(series_test.start_time() == pd.to_datetime('20130101'))
        self.assertTrue(series_test.end_time() == pd.to_datetime('20130110'))
        self.assertTrue(series_test.pd_series().equals(self.pd_series1))
        self.assertTrue(series_test.conf_lo_pd_series().equals(self.pd_series2))
        self.assertTrue(series_test.conf_hi_pd_series().equals(self.pd_series3))
        self.assertTrue(series_test.freq() == self.series1.freq())

    # TODO test over to_dataframe when multiple features choice is decided

    def test_eq(self):
        seriesA: TimeSeries = TimeSeries(self.pd_series1)
        self.assertTrue(self.series1 == seriesA)

        # with a defined CI
        seriesB: TimeSeries = TimeSeries(self.pd_series1,
                                         confidence_hi=pd.Series(range(10, 20),
                                                                 index=pd.date_range('20130101', '20130110')))
        self.assertFalse(self.series1 == seriesB)
        self.assertTrue(self.series1 != seriesB)

        # with different dates
        seriesC = TimeSeries(pd.Series(range(10), index=pd.date_range('20130102', '20130111')))
        self.assertFalse(self.series1 == seriesC)

        # compare with both CI
        seriesD: TimeSeries = TimeSeries(self.pd_series1, self.pd_series2, self.pd_series3)
        seriesE: TimeSeries = TimeSeries(self.pd_series1, self.pd_series3, self.pd_series2)
        self.assertTrue(self.series2 == seriesD)
        self.assertFalse(self.series2 == seriesE)

    def test_dates(self):
        self.assertEqual(self.series1.start_time(), pd.Timestamp('20130101'))
        self.assertEqual(self.series1.end_time(), pd.Timestamp('20130110'))
        self.assertEqual(self.series1.duration(), pd.Timedelta(days=9))

    def test_slice(self):
        # base case
        seriesA = self.series1.slice(pd.Timestamp('20130104'), pd.Timestamp('20130107'))
        self.assertEqual(seriesA.start_time(), pd.Timestamp('20130104'))
        self.assertEqual(seriesA.end_time(), pd.Timestamp('20130107'))

        # time stamp not in series
        seriesB = self.series1.slice(pd.Timestamp('20130104 12:00:00'), pd.Timestamp('20130107'))
        self.assertEqual(seriesB.start_time(), pd.Timestamp('20130105'))
        self.assertEqual(seriesB.end_time(), pd.Timestamp('20130107'))

        # end timestamp after series
        seriesC = self.series1.slice(pd.Timestamp('20130108'), pd.Timestamp('20130201'))
        self.assertEqual(seriesC.start_time(), pd.Timestamp('20130108'))
        self.assertEqual(seriesC.end_time(), pd.Timestamp('20130110'))

        # n points, base case
        seriesD = self.series1.slice_n_points_after(pd.Timestamp('20130102'), n=3)
        self.assertEqual(seriesD.start_time(), pd.Timestamp('20130102'))
        self.assertTrue(len(seriesD.values()) == 3)
        self.assertEqual(seriesD.end_time(), pd.Timestamp('20130104'))

        seriesE = self.series1.slice_n_points_after(pd.Timestamp('20130107 12:00:10'), n=10)
        self.assertEqual(seriesE.start_time(), pd.Timestamp('20130108'))
        self.assertEqual(seriesE.end_time(), pd.Timestamp('20130110'))

        seriesF = self.series1.slice_n_points_before(pd.Timestamp('20130105'), n=3)
        self.assertEqual(seriesF.end_time(), pd.Timestamp('20130105'))
        self.assertTrue(len(seriesF.values()) == 3)
        self.assertEqual(seriesF.start_time(), pd.Timestamp('20130103'))

        seriesG = self.series1.slice_n_points_before(pd.Timestamp('20130107 12:00:10'), n=10)
        self.assertEqual(seriesG.start_time(), pd.Timestamp('20130101'))
        self.assertEqual(seriesG.end_time(), pd.Timestamp('20130107'))

        # with CI
        seriesH = self.series2.slice(pd.Timestamp('20130104'), pd.Timestamp('20130107'))
        self.assertEqual(seriesH.conf_lo_pd_series().index[0], pd.Timestamp('20130104'))
        self.assertEqual(seriesH.conf_lo_pd_series().index[-1], pd.Timestamp('20130107'))
        self.assertEqual(seriesH.conf_hi_pd_series().index[0], pd.Timestamp('20130104'))
        self.assertEqual(seriesH.conf_hi_pd_series().index[-1], pd.Timestamp('20130107'))

    def test_split(self):
        seriesA, seriesB = self.series1.split_after(pd.Timestamp('20130104'))
        self.assertEqual(seriesA.end_time(), pd.Timestamp('20130104'))
        self.assertEqual(seriesB.start_time(), pd.Timestamp('20130105'))

        seriesC, seriesD = self.series1.split_before(pd.Timestamp('20130104'))
        self.assertEqual(seriesC.end_time(), pd.Timestamp('20130103'))
        self.assertEqual(seriesD.start_time(), pd.Timestamp('20130104'))

        self.assertEqual(self.series1.freq_str(), seriesA.freq_str())
        self.assertEqual(self.series1.freq_str(), seriesC.freq_str())

    def test_drop(self):
        seriesA = self.series1.drop_after(pd.Timestamp('20130105'))
        self.assertEqual(seriesA.end_time(), pd.Timestamp('20130105') - self.series1.freq())
        self.assertTrue(np.all(seriesA.time_index() < pd.Timestamp('20130105')))

        seriesB = self.series1.drop_before(pd.Timestamp('20130105'))
        self.assertEqual(seriesB.start_time(), pd.Timestamp('20130105') + self.series1.freq())
        self.assertTrue(np.all(seriesB.time_index() > pd.Timestamp('20130105')))

        self.assertEqual(self.series1.freq_str(), seriesA.freq_str())
        self.assertEqual(self.series1.freq_str(), seriesB.freq_str())

    def test_intersect(self):
        seriesA = TimeSeries(pd.Series(range(2, 8), index=pd.date_range('20130102', '20130107')))

        seriesB = self.series1.intersect(seriesA)
        self.assertEqual(seriesB.start_time(), pd.Timestamp('20130102'))
        self.assertEqual(seriesB.end_time(), pd.Timestamp('20130107'))

        # The same, with CI
        seriesC = self.series2.intersect(seriesA)
        self.assertEqual(seriesC.conf_lo_pd_series().index[0], pd.Timestamp('20130102'))
        self.assertEqual(seriesC.conf_hi_pd_series().index[-1], pd.Timestamp('20130107'))

        # Outside of range
        seriesD = self.series1.intersect(TimeSeries(pd.Series(range(6, 13),
                                                              index=pd.date_range('20130106', '20130112'))))
        self.assertEqual(seriesD.start_time(), pd.Timestamp('20130106'))
        self.assertEqual(seriesD.end_time(), pd.Timestamp('20130110'))

        # No intersect or too small intersect
        with self.assertRaises(AssertionError):
            self.series1.intersect(TimeSeries(pd.Series(range(6, 13),
                                                        index=pd.date_range('20130116', '20130122'))))
        with self.assertRaises(AssertionError):
            self.series1.intersect(TimeSeries(pd.Series(range(9, 13),
                                                        index=pd.date_range('20130109', '20130112'))))

    def test_rescale(self):
        with self.assertRaises(AssertionError):
            self.series1.rescale_with_value(1)

        seriesA = self.series3.rescale_with_value(0)
        self.assertTrue(np.all(seriesA.values() == 0))

        seriesB = self.series3.rescale_with_value(-5)
        self.assertTrue(self.series3 * -1. == seriesB)

        seriesC = self.series3.rescale_with_value(1)
        self.assertTrue(self.series3 * 0.2 == seriesC)

        seriesD = self.series3.rescale_with_value(1e+20)  # TODO: test will fail if value > 1e24 due to num imprecision
        self.assertTrue(self.series3 * 0.2e+20 == seriesD)

    def test_shift(self):
        seriesA = self.series1.shift(0)
        self.assertTrue(seriesA == self.series1)

        seriesB = self.series1.shift(1)
        self.assertTrue(seriesB.time_index().equals(
                        self.series1.time_index()[1:].append(pd.DatetimeIndex([self.series1.time_index()[-1] +
                                                                              self.series1.freq()]))))

        seriesC = self.series1.shift(-1)
        self.assertTrue(seriesC.time_index().equals(
            pd.DatetimeIndex([self.series1.time_index()[0] - self.series1.freq()]).append(
                self.series1.time_index()[:-1])))

        with self.assertRaises(OverflowError):
            self.series1.shift(1e+6)

        seriesM = TimeSeries.from_times_and_values(pd.date_range('20130101', '20130601', freq='m'), range(5))
        with self.assertRaises(OverflowError):
            seriesM.shift(1e+4)

    def test_append(self):
        # reconstruct series
        seriesA, seriesB = self.series1.split_after(pd.Timestamp('20130106'))
        self.assertEqual(seriesA.append(seriesB), self.series1)
        self.assertEqual(seriesA.append(seriesB).freq(), self.series1.freq())

        # Creating a gap is not allowed
        seriesC = self.series1.drop_before(pd.Timestamp('20130107'))
        with self.assertRaises(AssertionError):
            seriesA.append(seriesC)

        # Changing frequence is not allowed
        seriesM = TimeSeries.from_times_and_values(pd.date_range('20130107', '20130507', freq='30D'), range(5))
        with self.assertRaises(AssertionError):
            seriesA.append(seriesM)

        # reconstruction with CI
        seriesD, seriesE = self.series2.split_after(pd.Timestamp('20130106'))
        self.assertEqual(seriesD.append(seriesE), self.series2)
        self.assertEqual(seriesD.append(seriesE).freq(), self.series2.freq())

    def test_append_values(self):
        # reconstruct series
        seriesA, seriesB = self.series1.split_after(pd.Timestamp('20130106'))
        self.assertEqual(seriesA.append_values(seriesB.values(), seriesB.time_index()), self.series1)
        self.assertEqual(seriesA.append_values(seriesB.values()), self.series1)

        # same with CI
        seriesC, seriesD = self.series2.split_after(pd.Timestamp('20130106'))
        self.assertEqual(seriesC.append_values(seriesD.values(), seriesD.time_index(),
                                     seriesD.conf_lo_pd_series().values,
                                     seriesD.conf_hi_pd_series().values), self.series2)

        # add only few element
        self.assertEqual(self.series1.drop_after(pd.Timestamp('20130110')).append_values([9]), self.series1)
        self.assertEqual(seriesA.append_values([]), seriesA)

        # randomize order
        rd_order = np.random.permutation(range(len(seriesB.values())))
        self.assertEqual(seriesA.append_values(seriesB.values()[rd_order], seriesB.time_index()[rd_order]),
                         self.series1)

        # add non consecutive index
        with self.assertRaises(AssertionError):
            self.assertEqual(seriesA.append_values(seriesB.values(), seriesB.time_index()+seriesB.freq()), self.series1)

        # add existing indices
        with self.assertRaises(AssertionError):
            self.assertEqual(seriesA.append_values(seriesB.values(), seriesB.time_index()-3*seriesB.freq()), self.series1)

        # other frequency
        with self.assertRaises(AssertionError):
            self.assertEqual(seriesA.append_values(seriesB.values(), pd.date_range('20130107', '20130113', freq='2d')),
                             self.series1)

    def test_update(self):
        seriesA: TimeSeries = TimeSeries.from_times_and_values(self.times, [0, 1, 1, 3, 4, 5, 6, 2, 8, 0])
        seriesB: TimeSeries = TimeSeries.from_times_and_values(self.times, range(10),
                                                               [5, 1, 7, 3, 9, 5, 11, 2, 13, 14],
                                                               [15, 16, 1, 18, 4, 20, 6, 22, 8, 24])
        # change nothing
        seriesC = self.series1.copy()
        with self.assertRaises(AssertionError):
            seriesA.update(self.times)
        seriesC.update(self.times, range(10))
        self.assertEqual(seriesC, self.series1)

        # different len
        with self.assertRaises(AssertionError):
            seriesA.update(self.times, [], None, None)
        with self.assertRaises(AssertionError):
            seriesA.update(self.times, None, np.arange(3), None)
        with self.assertRaises(AssertionError):
            seriesA.update(self.times, None, None, np.arange(4))

        # change outside
        seriesC = seriesA.copy()
        with self.assertRaises(AssertionError):
            seriesC.update(self.times+100*seriesC.freq(), range(10))
        seriesC.update(self.times.append(pd.date_range('20140101', '20140110')), list(range(10))+[0]*10)
        self.assertEqual(seriesC, self.series1)

        # change random
        seriesC = seriesA.copy()
        seriesC.update(pd.DatetimeIndex(['20130108', '20130110', '20130103']), [7, 9, 2])
        self.assertEqual(seriesC, self.series1)

        # change one of each series
        seriesD = seriesB.copy()
        seriesD.update(self.times, seriesA.pd_series())
        seriesA.update(pd.DatetimeIndex(['20130103', '20130108', '20130110']), [2, 7, 9])
        self.assertEqual(seriesA, self.series1)
        seriesB.update(self.times[::2], conf_hi=range(15, 25, 2))
        self.assertTrue(seriesB.conf_hi_pd_series().equals(self.series2.conf_hi_pd_series()))
        self.assertNotEqual(seriesB, self.series2)
        seriesB.update(self.times[1::2], conf_lo=range(6, 15, 2))
        self.assertEqual(seriesB, self.series2)

        # use nan to update all series altogether
        new_series = np.empty(10)
        new_series[:] = np.nan
        new_series[[2, 7, 9]] = [2, 7, 9]
        new_lo = np.empty(10)
        new_lo[:] = np.nan
        new_lo[1::2] = np.arange(6, 15, 2)
        new_hi = np.empty(10)
        new_hi[:] = np.nan
        new_hi[::2] = np.arange(15, 25, 2)
        seriesD.update(self.times, new_series, new_lo, new_hi)
        self.assertEqual(seriesD, self.series2)

        # raise error when update missing CI
        with self.assertRaises(AttributeError):
            self.series1.update(self.times, conf_lo=range(5, 15))

    def test_drop_values(self):
        seriesA = self.series1.append_values([1])
        self.assertEqual(seriesA.drop_values(pd.Timestamp('20130111'), inplace=False), self.series1)
        seriesA.drop_values(pd.Timestamp('20130111'))
        self.assertEqual(seriesA, self.series1)

        with self.assertRaises(KeyError):
            seriesA.drop_values(pd.Timestamp('20130112'))

    def test_ops(self):
        seriesA = TimeSeries(pd.Series([2 for _ in range(10)], index=self.pd_series1.index))
        targetAdd = TimeSeries(pd.Series(range(2, 12), index=self.pd_series1.index))
        targetSub = TimeSeries(pd.Series(range(-2, 8), index=self.pd_series1.index))
        targetMul = TimeSeries(pd.Series(range(0, 20, 2), index=self.pd_series1.index))
        targetDiv = TimeSeries(pd.Series([i / 2 for i in range(10)], index=self.pd_series1.index))
        targetPow = TimeSeries(pd.Series([float(i ** 2) for i in range(10)], index=self.pd_series1.index))

        self.assertEqual(self.series1 + seriesA, targetAdd)
        self.assertEqual(self.series1 + 2, targetAdd)
        self.assertEqual(2 + self.series1, targetAdd)
        self.assertEqual(self.series1 - seriesA, targetSub)
        self.assertEqual(self.series1 - 2, targetSub)
        self.assertEqual(self.series1 * seriesA, targetMul)
        self.assertEqual(self.series1 * 2, targetMul)
        self.assertEqual(2 * self.series1, targetMul)
        self.assertEqual(self.series1 / seriesA, targetDiv)
        self.assertEqual(self.series1 / 2, targetDiv)
        self.assertEqual(self.series1 ** 2, targetPow)

        with self.assertRaises(AssertionError):
            # Cannot divide by a TimeSeries with a value 0.
            self.series1 / self.series1

        with self.assertRaises(AssertionError):
            # Cannot divide by 0.
            self.series1 / 0

    def test_getitem(self):
        seriesA: TimeSeries = self.series1.drop_after(pd.Timestamp("20130105"))
        self.assertEqual(self.series1[pd.date_range('20130101', ' 20130104')], seriesA)
        self.assertEqual(self.series1[:4], seriesA)
        self.assertTrue(self.series1[pd.Timestamp('20130101')].equals(self.series1.pd_series()[:1]))
        self.assertEqual(self.series1[pd.Timestamp('20130101'):pd.Timestamp('20130105')], seriesA)

        with self.assertRaises(IndexError):
            self.series1[pd.date_range('19990101', '19990201')]

        with self.assertRaises(IndexError):
            self.series1['19990101']

        with self.assertRaises(IndexError):
            self.series1[::-1]





from setuptools import setup, find_packages

setup(name='u8timeseries',
      version='0.0.1',
      description='A collection of easy-to-use timeseries forecasting models',
      url='http://github.com/unit8co/u8timeseries',
      author='Unit8 SA',
      author_email='info@unit8.co',
      license='Apache License 2.0',
      packages=find_packages(),
      install_requires=[
          'numpy>=1.15.4',
          'dateutils',
          'scipy<1.3',
          'statsmodels==0.9.0',
          'pmdarima<1.3',
          'plotly',
          'fbprophet>=0.5',
          'scikit-learn>=0.21.2',
          'pandas>=0.23.1',
          'tqdm>=4.32.1',
          'ipywidgets'
      ],
      zip_safe=False,
      python_requires='>=3.6'
      )



import numpy as np
from typing import Tuple
from u8timeseries.timeseries import TimeSeries
from warnings import warn


def _import_check_seasonality():
    try:
        from u8timeseries.models.statistics import check_seasonality as cs
    except ImportError as e:
        raise ImportError('Cannot import check_seasonality. Choose a fixed period')
    return cs


def _get_values_or_raise(series_a: TimeSeries, series_b: TimeSeries) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns the numpy values of two time series, launching an Exception if time series cannot be compared
    """
    assert series_a.has_same_time_as(series_b), 'The two time series must have same time index.' \
                                                '\nFirst series: {}\nSecond series: {}'.format(
                                                series_a.time_index(), series_b.time_index())
    return series_a.values(), series_b.values()


def mae(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Compute the Mean Absolute Error (MAE).

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the MAE of `pred_series` with respect to `true_series`.
    """
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_pred = np.diff(y_true), np.diff(y_pred)
    return np.mean(np.abs(y_true - y_pred))


def mse(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Compute the Mean Squared Error (MSE).

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the MSE of `pred_series` with respect to `true_series`.
    """
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_pred = np.diff(y_true), np.diff(y_pred)
    return np.mean((y_true - y_pred)**2)


def rmse(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Compute the Root Mean Squared Error (RMSE).

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the RMSE of `pred_series` with respect to `true_series`.
    """
    return np.sqrt(mse(true_series, pred_series, time_diff))


def rmsle(true_series: TimeSeries, pred_series: TimeSeries) -> float:
    """
    Compute the Root Mean Squared Log Error (RMSLE).

    Penalize more the under-estimate than the over-estimate.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :return: A float, the RMSLE of `pred_series` with respect to `true_series`.
    """
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    y_true, y_pred = np.log(y_true + 1), np.log(y_pred + 1)
    return np.sqrt(np.mean((y_true - y_pred)**2))


def coefficient_variation(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Compute the Root Mean Squared Error (RMSE).

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the RMSE of `pred_series` with respect to `true_series`.
    """
    return 100 * rmse(true_series, pred_series, time_diff)/true_series.mean()


def mape(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Computes the Mean Absolute Percentage Error (MAPE).

    This function computes the MAPE of `pred_series` with respect to `true_series`.
    Use `time_diff=True` when the time series has a strong auto-correlation to have a more accurate analysis.
    Use `time_diff` mainly when the time series are monotonic.
    Otherwise, it will be difficult to analyze the results.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the MAPE of `pred_series` with respect to `true_series`.
    """
    # Mean absolute percentage error
    y_true, y_hat = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_hat = np.diff(y_true), np.diff(y_hat)
    return 100. * np.mean(np.abs((y_true - y_hat) / y_true))


def mase_old(true_series: TimeSeries, pred_series: TimeSeries) -> float:
    """
    Computes the Mean Absolute Scaled Error (MASE).

    This function computes the MASE of `pred_series` with respect to `true_series`.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :return: A float, the MASE of `pred_series` with respect to `true_series`.
    """
    warn("This function does not take into account the seasonality of the time series. "
         "Please use mase_seasonal to be accurate", FutureWarning)
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    errors = np.sum(np.abs(y_true - y_pred))
    t = y_true.size
    scale = t/(t-1) * np.sum(np.abs(np.diff(y_true)))
    return errors / scale


def mase(true_series: TimeSeries, pred_series: TimeSeries, m: int = 1, time_diff: bool = False) -> float:
    """
    Computes the Mean Absolute Scaled Error (MASE).

    This function computes the MASE of `pred_series` with respect to `true_series` and the seasonal period m.
    Use `time_diff=True` when the time series has a strong auto-correlation to have a more accurate analysis.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param m: A int, the seasonality period to take into account. If None, try to infer one from ACF
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the MASE of `pred_series` with respect to `true_series`.
    """
    if m is None:
        check_seasonality = _import_check_seasonality()
        test_season, m = check_seasonality()
        if not test_season:
            warn("No seasonality found. The period is fixed to 1.", UserWarning)
            m = 1
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_pred = np.diff(y_true), np.diff(y_pred)
    errors = np.sum(np.abs(y_true - y_pred))
    t = y_true.size
    scale = t/(t-m) * np.sum(np.abs(y_true[m:] - y_true[:-m]))
    assert not np.isclose(scale, 0), "cannot use MASE with periodical signals"
    return errors / scale


def overall_percentage_error(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Computes the Overall Percentage Erroe (OPE):

    This function computes the OPE of `pred_series` with respect to `true_series`.
    Use `time_diff=True` when the time series has a strong auto-correlation to have a more accurate analysis.
    Use `time_diff` mainly when the time series are monotonic.
    Otherwise, it will be difficult to analyze the results.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the OPE of `pred_series` with respect to `true_series`.
    """
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_pred = np.diff(y_true), np.diff(y_pred)
    y_true_sum, y_pred_sum = np.sum(np.array(y_true)), np.sum(np.array(y_pred))
    return np.abs((y_true_sum - y_pred_sum) / y_true_sum) * 100.


def marre(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Computes the Mean Absolute Ranged Relative Error (MARRE).

    This function computes the MARRE of `pred_series` with respect to `true_series`.
    Use `time_diff=True` when the time series has a strong auto-correlation to have a more accurate analysis.

    :param true_series: A TimeSeries.
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the MARRE of `pred_series` with respect to `true_series`.
    """
    y_true, y_hat = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_hat= np.diff(y_true), np.diff(y_hat)
    true_range = y_true.max() - y_true.min()
    return 100. * np.mean(np.abs((y_true - y_hat) / true_range))


def r2_score(true_series: TimeSeries, pred_series: TimeSeries, time_diff: bool = False) -> float:
    """
    Computes the coefficient of determination R2.
    Use `time_diff=True` when the time series has a strong auto-correlation to have a more accurate analysis.

    This function computes the R2 score of `pred_series` with respect to `true_series`.
    :param true_series: A TimeSeries
    :param pred_series: A TimeSeries to be compared with `true_series`.
    :param time_diff: If True, analyze the time differentiated series, instead of the index one.
    :return: A float, the coefficient R2
    """
    y_true, y_pred = _get_values_or_raise(true_series, pred_series)
    if time_diff:
        y_true, y_pred = np.diff(y_true), np.diff(y_pred)
    ss_errors = np.sum((y_true - y_pred)**2)
    y_hat = y_true.mean()
    ss_tot = np.sum((y_true-y_hat)**2)
    return 1 - ss_errors/ss_tot

